#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Provides routines that support SWPC space weather instruments."""

import datetime as dt
import ftplib
import numpy as np
import os
import pandas as pds
import sys

import pysat

from pysatSpaceWeather.instruments.methods import general

# ----------------------------------------------------------------------------
# Define the module variables

ackn = ''.join(['Prepared by the U.S. Dept. of Commerce, NOAA, Space ',
                'Weather Prediction Center'])
forecast_warning = ''.join(['This routine can only download the current ',
                            'forecast, not archived forecasts'])


# ----------------------------------------------------------------------------
# Define the module functions

def daily_dsd_download(name, today, data_path, mock_download_dir=None):
    """Download the daily NOAA Daily Solar Data indices.

    Parameters
    ----------
    name : str
        Instrument name, expects one of 'f107', 'flare', 'ssn', or 'sbfield'.
    today : dt.datetime
        Datetime for current day.
    data_path : str
        Path to data directory.
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded (default=None)

    Raises
    ------
    IOError
        If an unknown mock download directory is supplied or the desired file
        is missing.

    Note
    ----
    Note that the download path for the complementary Instrument will use
    the standard pysat data paths

    """
    pysat.logger.info('This routine only downloads the latest 30 day file')

    # Get the file information
    raw_txt = general.get_local_or_remote_text(
        'https://services.swpc.noaa.gov/text/', mock_download_dir,
        'daily-solar-indices.txt')

    if raw_txt is None:
        pysat.logger.info("".join(["Data not downloaded for ",
                                   "daily-solar-indices.txt, data may have ",
                                   "been saved to an unexpected filename."]))
    else:
        # Get the file paths and output names
        file_paths = {data_name: data_path if name == data_name else
                      general.get_instrument_data_path(
                          'sw_{:s}'.format(data_name), tag='daily')
                      for data_name in ['f107', 'flare', 'ssn', 'sbfield']}
        outfiles = {
            data_name: os.path.join(file_paths[data_name], '_'.join([
                data_name, 'daily', '{:s}.txt'.format(
                    today.strftime('%Y-%m-%d'))]))
            for data_name in file_paths.keys()}

        # Check that the directories exist
        for data_path in file_paths.values():
            pysat.utils.files.check_and_make_path(data_path)

        # Save the output
        rewrite_daily_solar_data_file(today.year, outfiles, raw_txt)

    return


def old_indices_dsd_download(name, date_array, data_path, local_files, today,
                             mock_download_dir=None):
    """Download the old NOAA Daily Solar Data indices.

    Parameters
    ----------
    name : str
        Instrument name, expects one of 'f107', 'flare', 'ssn', or 'sbfield'.
    date_array : array-like or pandas.DatetimeIndex
        Array-like or index of datetimes to be downloaded.
    data_path : str
        Path to data directory.
    local_files : pds.Series
        A Series containing the local filenames indexed by time.
    today : dt.datetime
        Datetime for current day
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded (default=None)

    Raises
    ------
    IOError
        If an unknown mock download directory is supplied.

    Note
    ----
    Note that the download path for the complementary Instrument will use
    the standard pysat data paths

    """
    # Get the file paths
    file_paths = {data_name: data_path if name == data_name else
                  general.get_instrument_data_path('sw_{:s}'.format(data_name),
                                                   tag='prelim')
                  for data_name in ['f107', 'flare', 'ssn', 'sbfield']}

    # Check that the directories exist
    for data_path in file_paths.values():
        pysat.utils.files.check_and_make_path(data_path)

    if mock_download_dir is None:
        # Connect to the host, default port
        ftp = ftplib.FTP('ftp.swpc.noaa.gov')
        ftp.login()  # User anonymous, passwd anonymous
        ftp.cwd('/pub/indices/old_indices')
    elif not os.path.isdir(mock_download_dir):
        raise IOError('file location is not a directory: {:}'.format(
            mock_download_dir))

    bad_fname = list()

    # To avoid downloading multiple files, cycle dates based on file length
    dl_date = date_array[0]
    while dl_date <= date_array[-1]:
        # The file name changes, depending on how recent the requested data is
        qnum = (dl_date.month - 1) // 3 + 1  # Integer floor division
        qmonth = (qnum - 1) * 3 + 1
        quar = 'Q{:d}_'.format(qnum)
        fnames = ['{:04d}{:s}DSD.txt'.format(dl_date.year, ss)
                  for ss in ['_', quar]]
        versions = ["01_v2", "{:02d}_v1".format(qmonth)]
        vend = [dt.datetime(dl_date.year, 12, 31),
                dt.datetime(dl_date.year, qmonth, 1)
                + pds.DateOffset(months=3) - pds.DateOffset(days=1)]
        downloaded = False
        rewritten = False

        # Attempt the download(s)
        for iname, fname in enumerate(fnames):
            # Test to see if we already tried this filename
            if fname in bad_fname:
                continue

            local_fname = fname
            saved_fname = os.path.join(data_path, local_fname)
            outfiles = {
                data_name: os.path.join(file_paths[data_name], '_'.join(
                    [data_name, 'prelim', '{:04d}'.format(dl_date.year),
                     '{:s}.txt'.format(versions[iname])]))
                for data_name in file_paths.keys()}

            if os.path.isfile(outfiles[name]):
                downloaded = True

                # Check the date to see if this should be rewritten
                checkfile = os.path.split(outfiles[name])[-1]
                has_file = local_files == checkfile
                if np.any(has_file):
                    if has_file[has_file].index[-1] < vend[iname]:
                        # This file will be updated again, but only attempt to
                        # do so if enough time has passed from the last time it
                        # was downloaded
                        yesterday = today - pds.DateOffset(days=1)
                        if has_file[has_file].index[-1] < yesterday:
                            rewritten = True
            else:
                # The file does not exist, if it can be downloaded, it
                # should be 'rewritten'
                rewritten = True

            # Attempt to download if the file does not exist or if the
            # file has been updated
            if rewritten or not downloaded:
                if mock_download_dir is None:
                    try:
                        sys.stdout.flush()
                        ftp.retrbinary('RETR ' + fname,
                                       open(saved_fname, 'wb').write)
                        downloaded = True
                        pysat.logger.info(' '.join(('Downloaded file for ',
                                                    dl_date.strftime('%x'))))

                    except ftplib.error_perm as exception:
                        # Could not fetch, so cannot rewrite
                        rewritten = False

                        # Test for an error
                        if str(exception.args[0]).split(" ", 1)[0] != '550':
                            raise IOError(exception)
                        else:
                            # File isn't actually there, try the next name.  The
                            # extra wrapping is for Windows, which can encounter
                            # permission errors when handling files.
                            attempt = 0
                            while attempt < 100:
                                try:
                                    os.remove(saved_fname)
                                    attempt = 100
                                except PermissionError:
                                    attempt += 1

                            # Save this so we don't try again. Because there are
                            # two possible filenames for each time, it's ok if
                            # one isn't there.  We just don't want to keep
                            # looking for it.
                            bad_fname.append(fname)
                else:
                    # Set the saved filename
                    saved_fname = os.path.join(mock_download_dir, local_fname)
                    downloaded = True

                    if os.path.isfile(saved_fname):
                        rewritten = True
                    else:
                        pysat.logger.info("".join([saved_fname, "is missing, ",
                                                   "data may have been saved ",
                                                   "to an unexpected ",
                                                   "filename."]))
                        rewritten = False

            # If the first file worked, don't try again
            if downloaded:
                break

        if not downloaded:
            pysat.logger.info(' '.join(('File not available for',
                                        dl_date.strftime('%x'))))
        elif rewritten:
            with open(saved_fname, 'r') as fprelim:
                lines = fprelim.read()

            rewrite_daily_solar_data_file(dl_date.year, outfiles, lines)
            if mock_download_dir is None:
                # Only remove the file if it wasn't obtained from a local dir
                os.remove(saved_fname)

        # Cycle to the next date
        dl_date = vend[iname] + pds.DateOffset(days=1)

    # Close connection after downloading all dates
    if mock_download_dir is None:
        ftp.close()
    return


def rewrite_daily_solar_data_file(year, outfiles, lines):
    """Rewrite the SWPC Daily Solar Data files.

    Parameters
    ----------
    year : int
        Year of data file (format changes based on date)
    outfiles : dict
        Output filenames for all relevant Instruments
    lines : str
        String containing all output data (result of 'read')

    """

    # Get to the solar index data
    if year > 2000:
        raw_data = lines.split('#---------------------------------')[-1]
        raw_data = raw_data.split('\n')[1:-1]
        optical = True
    else:
        raw_data = lines.split('# ')[-1]
        raw_data = raw_data.split('\n')
        optical = False if raw_data[0].find('Not Available') or year == 1994 \
            else True
        istart = 7 if year < 2000 else 1
        raw_data = raw_data[istart:-1]

    # Parse the data
    solar_times, data_dict = parse_daily_solar_data(raw_data, year, optical)

    # Separate data by Instrument name
    data_cols = {'f107': ['f107'],
                 'flare': ['goes_bgd_flux', 'c_flare', 'm_flare', 'x_flare',
                           'o1_flare', 'o2_flare', 'o3_flare', 'o1_flare',
                           'o2_flare', 'o3_flare', 'c_flare', 'm_flare',
                           'x_flare'],
                 'ssn': ['ssn', 'ss_area', 'new_reg'],
                 'sbfield': ['smf']}

    for data_name in data_cols.keys():
        name_dict = {dkey: data_dict[dkey] for dkey in data_dict.keys()
                     if dkey in data_cols[data_name]}

        # Collect into DataFrame
        data = pds.DataFrame(name_dict, index=solar_times)

        # Write out as a file
        data.to_csv(outfiles[data_name], header=True)

    return


def parse_daily_solar_data(data_lines, year, optical):
    """Parse the data in the SWPC daily solar index file.

    Parameters
    ----------
    data_lines : list
        List of lines containing data
    year : list
        Year of file
    optical : boolean
        Flag denoting whether or not optical data is available

    Returns
    -------
    dates : list
        List of dates for each date/data pair in this block
    values : dict
        Dict of lists of values, where each key is the value name

    """

    # Initialize the output
    dates = list()
    val_keys = ['f107', 'ssn', 'ss_area', 'new_reg', 'smf', 'goes_bgd_flux',
                'c_flare', 'm_flare', 'x_flare', 'o1_flare', 'o2_flare',
                'o3_flare']
    optical_keys = ['o1_flare', 'o2_flare', 'o3_flare']
    xray_keys = ['c_flare', 'm_flare', 'x_flare']
    values = {kk: list() for kk in val_keys}

    # Cycle through each line in this file
    for line in data_lines:
        # Split the line on whitespace
        split_line = line.split()

        # Format the date
        dfmt = "%Y %m %d" if year > 1996 else "%d %b %y"
        dates.append(dt.datetime.strptime(" ".join(split_line[0:3]), dfmt))

        # Format the data values
        j = 0
        for i, kk in enumerate(val_keys):
            if year == 1994 and kk == 'new_reg':
                # New regions only in files after 1994
                val = -999
            elif np.any([year == 1994 and kk in xray_keys,
                         not optical and kk in optical_keys]):
                # X-ray flares in files after 1994, optical flares come later
                val = -1
            else:
                val = split_line[j + 3]
                j += 1

            if kk != 'goes_bgd_flux':
                if val == "*":
                    val = -999 if i < 5 else -1
                else:
                    val = int(val)
            values[kk].append(val)

    return dates, values


def solar_geomag_predictions_download(name, date_array, data_path,
                                      mock_download_dir=None):
    """Download the 3-day solar-geomagnetic predictions from SWPC.

    Parameters
    ----------
    name : str
        Instrument name, expects one of 'kp', 'ap', 'stormprob', 'f107',
        'flare', or 'polarcap'.
    date_array : array-like or pandas.DatetimeIndex
        Array-like or index of datetimes to be downloaded.
    data_path : str
        Path to data directory.
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded (default=None)

    Raises
    ------
    IOError
        If an unknown mock download directory is supplied or the desired file
        is missing.

    Note
    ----
    Note that the download path for the complementary Instrument will use
    the standard pysat data paths

    """
    pysat.logger.info(forecast_warning)

    # Get the file paths
    file_paths = {data_name: data_path if name == data_name else
                  general.get_instrument_data_path(
                      'sw_{:s}'.format(data_name),
                      tag='forecast' if data_name == 'f107' else 'prediction')
                  for data_name in ['kp', 'ap', 'stormprob', 'f107', 'flare',
                                    'polarcap']}

    # Check that the directories exist
    for data_path in file_paths.values():
        pysat.utils.files.check_and_make_path(data_path)

    # Get the file information
    raw_txt = general.get_local_or_remote_text(
        'https://services.swpc.noaa.gov/text/', mock_download_dir,
        '3-day-solar-geomag-predictions.txt')

    if raw_txt is None:
        pysat.logger.info("".join(["Data not downloaded for ",
                                   "3-day-solar-geomag-predictions.txt, data ",
                                   "may have been saved to an unexpected ",
                                   "filename."]))
    else:
        # Parse text to get the date the prediction was generated
        date_str = raw_txt.split(':Issued: ')[-1].split(' UTC')[0]
        dl_date = dt.datetime.strptime(date_str, '%Y %b %d %H%M')

        # Parse the data to get the prediction dates
        date_strs = raw_txt.split(':Prediction_dates:')[-1].split('\n')[0]
        pred_times = [
            dt.datetime.strptime(' '.join(date_str.split()), '%Y %b %d')
            for date_str in date_strs.split('  ') if len(date_str) > 0]

        # Separate out the data by chunks
        ap_raw = raw_txt.split(':Geomagnetic_A_indices:')[-1]
        kp_raw = raw_txt.split(':Pred_Mid_k:')[-1]
        storm_raw = raw_txt.split(':Prob_Mid:')[-1]
        pc_raw = raw_txt.split(':Polar_cap:')[-1]
        f107_raw = raw_txt.split(':10cm_flux:')[-1]
        flare_raw = raw_txt.split(':Whole_Disk_Flare_Prob:')[-1]

        # Initalize the data for each data type
        data_vals = {data_name: dict() for data_name in file_paths.keys()}
        data_times = {data_name: pred_times for data_name in file_paths.keys()}

        # Process the ap data
        for line in ap_raw.split('\n'):
            if line.find(":") == 0:
                break
            elif line.find("A_") == 0:
                split_line = line.split()
                if split_line[0] == "A_Planetary":
                    dkey = "daily_Ap"
                else:
                    dkey = split_line[0]

                data_vals['ap'][dkey] = [int(val) for val in split_line[1:]]

        # Process the Kp data
        hr_strs = ['00-03UT', '03-06UT', '06-09UT', '09-12UT', '12-15UT',
                   '15-18UT', '18-21UT', '21-00UT']
        data_times['kp'] = pds.date_range(pred_times[0], periods=24, freq='3H')

        for line in kp_raw.split('\n'):
            if line.find("Prob_Mid") >= 0:
                break
            elif line.find("UT") > 0:
                split_line = line.split()
                reg, hr = split_line[0].split('/')
                dkey = '{:s}_lat_Kp'.format(reg)

                # Initalize the Kp data for this region
                if dkey not in data_vals['kp'].keys():
                    data_vals['kp'][dkey] = np.full(shape=(24,),
                                                    fill_value=np.nan)

                # Save the Kp data into the correct day and hour index
                hr_index = hr_strs.index(hr)
                data_vals['kp'][dkey][hr_index] = float(split_line[1])
                data_vals['kp'][dkey][hr_index + 8] = float(split_line[2])
                data_vals['kp'][dkey][hr_index + 16] = float(split_line[3])

        # Process the storm probabilities
        for line in storm_raw.split('\n'):
            if line.find("Polar_cap") >= 0:
                break
            elif len(line) > 0:
                split_line = line.split()
                if split_line[0].find('/') > 0:
                    dkey = split_line[0].replace('/', '-Lat_')
                    data_vals['stormprob'][dkey] = [
                        int(val) for val in split_line[1:]]

        # Process the polar cap prediction
        data_vals['polarcap']['absorption_forecast'] = [
            str_val for str_val in pc_raw.split('\n')[1].split()]
        data_times['polarcap'] = [
            ptimes for i, ptimes in enumerate(pred_times)
            if i < len(data_vals['polarcap']['absorption_forecast'])]

        # Process the F10.7 data
        data_vals['f107']['f107'] = [
            int(val) for val in f107_raw.split('\n')[1].split()]

        # Process the flare data
        dkey_root = 'Whole_Disk_Flare_Prob'
        for line in flare_raw.split('\n'):
            if len(line) > 0 and line.find("#") < 0:
                if line.find(":") == 0:
                    dkey_root = line.split(":")[1]
                else:
                    split_line = line.split()

                    if len(split_line) == 4:
                        dkey = "_".join([dkey_root, split_line[0]])
                        data_vals['flare'][dkey] = [
                            int(val) for val in split_line[1:]]
                    else:
                        data_vals['flare']['{:s}_Region'.format(dkey_root)] = [
                            int(split_line[0]), -1, -1]
                        data_vals['flare']['{:s}_Class_C'.format(dkey_root)] = [
                            int(split_line[1]), -1, -1]
                        data_vals['flare']['{:s}_Class_M'.format(dkey_root)] = [
                            int(split_line[2]), -1, -1]
                        data_vals['flare']['{:s}_Class_X'.format(dkey_root)] = [
                            int(split_line[3]), -1, -1]
                        data_vals['flare']['{:s}_Class_P'.format(dkey_root)] = [
                            int(split_line[4]), -1, -1]

        # Save the data by type into files
        for data_name in data_vals.keys():
            # Put the data values into a nicer DataFrame
            data = pds.DataFrame(data_vals[data_name],
                                 index=data_times[data_name])

            # Save the data as a CSV file
            data_tag = 'forecast' if data_name == 'f107' else 'prediction'
            data_file = '_'.join([data_name, data_tag,
                                  '{:s}.txt'.format(dl_date.strftime(
                                      '%Y-%m-%d'))])
            data.to_csv(os.path.join(file_paths[data_name], data_file),
                        header=True)

    return


def geomag_forecast_download(name, date_array, data_path,
                             mock_download_dir=None):
    """Download the 3-day geomagnetic Kp, ap, and storm data from SWPC.

    Parameters
    ----------
    name : str
        Instrument name, expects one of 'kp', 'ap', or 'stormprob'.
    date_array : array-like or pandas.DatetimeIndex
        Array-like or index of datetimes to be downloaded.
    data_path : str
        Path to data directory.
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded (default=None)

    Raises
    ------
    IOError
        If an unknown mock download directory is supplied or the desired file
        is missing.

    Note
    ----
    Note that the download path for the complementary Instrument will use
    the standard pysat data paths

    """
    pysat.logger.info(forecast_warning)

    # Get the file paths
    file_paths = {data_name: data_path if name == data_name else
                  general.get_instrument_data_path(
                      'sw_{:s}'.format(data_name), tag='forecast')
                  for data_name in ['kp', 'ap', 'stormprob']}

    # Check that the directories exist
    for data_path in file_paths.values():
        pysat.utils.files.check_and_make_path(data_path)

    # Get the file information
    raw_txt = general.get_local_or_remote_text(
        'https://services.swpc.noaa.gov/text/', mock_download_dir,
        '3-day-geomag-forecast.txt')

    if raw_txt is None:
        pysat.logger.info("".join(["Data not downloaded for ",
                                   "3-day-geomag-forecast.txt, data may have ",
                                   "been saved to an unexpected filename."]))
    else:
        # Parse text to get the date the prediction was generated
        date_str = raw_txt.split(':Issued: ')[-1].split(' UTC')[0]
        dl_date = dt.datetime.strptime(date_str, '%Y %b %d %H%M')

        # Separate out the data by chunks
        ap_raw = raw_txt.split('NOAA Ap Index Forecast')[-1]
        kp_raw = raw_txt.split('NOAA Kp index forecast ')[-1]
        storm_raw = raw_txt.split('NOAA Geomagnetic Activity Probabilities')[-1]

        # Get dates of the forecasts
        date_str = kp_raw[0:6] + ' ' + str(dl_date.year)
        forecast_date = dt.datetime.strptime(date_str, '%d %b %Y')

        # Strings we will use to parse the downloaded text for Kp
        lines = ['00-03UT', '03-06UT', '06-09UT', '09-12UT', '12-15UT',
                 '15-18UT', '18-21UT', '21-00UT']

        # Storage for daily Kp forecasts. Get values for each day, then combine
        # them together
        kp_day1 = []
        kp_day2 = []
        kp_day3 = []
        for line in lines:
            raw = kp_raw.split(line)[-1].split('\n')[0]
            cols = raw.split()
            kp_day1.append(float(cols[-3]))
            kp_day2.append(float(cols[-2]))
            kp_day3.append(float(cols[-1]))

        kp_times = pds.date_range(forecast_date, periods=24, freq='3H')
        kp_day = []
        for dd in [kp_day1, kp_day2, kp_day3]:
            kp_day.extend(dd)

        # Put Kp data into nicer DataFrame
        data_frames = {'kp': pds.DataFrame(kp_day, index=kp_times,
                                           columns=['Kp'])}

        # Parse the Ap data
        ap_times = pds.date_range(dl_date - dt.timedelta(days=1), periods=5,
                                  freq='1D')
        obs_line = ap_raw.split('Observed Ap')[-1].split('\n')[0]
        est_line = ap_raw.split('Estimated Ap')[-1].split('\n')[0]
        pred_line = ap_raw.split('Predicted Ap')[-1].split('\n')[0]
        ap_vals = [int(obs_line[-3:]), int(est_line[-3:])]

        for ap_val in pred_line.split()[-1].split('-'):
            ap_vals.append(int(ap_val))

        # Put the Ap data into a nicer DataFrame
        data_frames['ap'] = pds.DataFrame(ap_vals, index=ap_times,
                                          columns=['daily_Ap'])

        # Parse the storm probabilities
        storm_dict = {}
        for storm_line in storm_raw.split('\n')[1:5]:
            storm_split = storm_line.split()

            # Build the storm data column name
            dkey = '_'.join(storm_split[:-1])

            # Assign the storm probabilities
            storm_dict[dkey] = [int(sp) for sp in storm_split[-1].split('/')]

        # Put the storm probabilities into a nicer DataFrame
        storm_times = pds.date_range(forecast_date, periods=3, freq='1D')
        data_frames['stormprob'] = pds.DataFrame(storm_dict, index=storm_times)

        # Save the data files
        for data_name in data_frames.keys():
            filename = '{:s}_forecast_{:s}.txt'.format(
                data_name, dl_date.strftime('%Y-%m-%d'))
            data_frames[data_name].to_csv(os.path.join(
                file_paths[data_name], filename), header=True)

    return


def kp_ap_recent_download(name, date_array, data_path, mock_download_dir=None):
    """Download recent Kp and ap data from SWPC.

    Parameters
    ----------
    name : str
        Instrument name, expects 'kp' or 'ap'.
    date_array : array-like or pandas.DatetimeIndex
        Array-like or index of datetimes to be downloaded.
    data_path : str
        Path to data directory.
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded (default=None)

    Raises
    ------
    IOError
        If an unknown mock download directory is supplied or the desired file
        is missing.

    Note
    ----
    Note that the download path for the complementary Instrument will use
    the standard pysat data paths

    """
    pysat.logger.info(forecast_warning)

    # Get the file paths
    file_paths = {data_name: data_path if name == data_name else
                  general.get_instrument_data_path(
                      'sw_{:s}'.format(data_name), tag='recent')
                  for data_name in ['kp', 'ap']}

    # Check that the directories exist
    for data_path in file_paths.values():
        pysat.utils.files.check_and_make_path(data_path)

    # Get the file information
    raw_txt = general.get_local_or_remote_text(
        'https://services.swpc.noaa.gov/text/', mock_download_dir,
        'daily-geomagnetic-indices.txt')

    if raw_txt is None:
        pysat.logger.info("".join(["Data not downloaded for ",
                                   "daily-geomagnetic-indices.txt, data may ",
                                   "have been saved to an unexpected ",
                                   "filename."]))
    else:
        # Parse text to get the date the prediction was generated
        date_str = raw_txt.split(':Issued: ')[-1].split('\n')[0]
        dl_date = dt.datetime.strptime(date_str, '%H%M UT %d %b %Y')

        # Data is the forecast value for the next three days
        raw_data = raw_txt.split('#  Date ')[-1]

        # Keep only the middle bits that matter
        raw_data = raw_data.split('\n')[1:-1]

        # Hold times from the file
        times = []

        # Holds Kp and Ap values for each station
        sub_kps = [[], [], []]
        sub_aps = [[], [], []]

        # Iterate through file lines and parse out the info we want
        for line in raw_data:
            times.append(dt.datetime.strptime(line[0:10], '%Y %m %d'))

            # Pick out Kp values for each of the three columns. The columns
            # used to all have integer values, but now some have floats.
            kp_sub_lines = [line[17:33], line[40:56], line[63:]]
            ap_sub_lines = [line[10:17], line[33:40], line[56:63]]
            for i, sub_line in enumerate(kp_sub_lines):
                # Process the Kp data, which has 3-hour values
                split_sub = sub_line.split()
                for ihr in np.arange(8):
                    if sub_line.find('.') < 0:
                        # These are integer values
                        sub_kps[i].append(
                            int(sub_line[(ihr * 2):((ihr + 1) * 2)]))
                    else:
                        # These are float values
                        sub_kps[i].append(np.float64(split_sub[ihr]))

                # Process the Ap data, which has daily values
                sub_aps[i].append(np.int64(ap_sub_lines[i]))

        # Create times on 3 hour cadence
        kp_times = pds.date_range(times[0], periods=(8 * 30), freq='3H')

        # Put both data sets into DataFrames
        data = {'kp': pds.DataFrame({'mid_lat_Kp': sub_kps[0],
                                     'high_lat_Kp': sub_kps[1],
                                     'Kp': sub_kps[2]}, index=kp_times),
                'ap': pds.DataFrame({'mid_lat_Ap': sub_aps[0],
                                     'high_lat_Ap': sub_aps[1],
                                     'daily_Ap': sub_aps[2]}, index=times)}

        # Write out the data sets as files
        for dkey in data.keys():
            data_file = '{:s}_recent_{:s}.txt'.format(
                dkey, dl_date.strftime('%Y-%m-%d'))

            data[dkey].to_csv(os.path.join(file_paths[dkey], data_file),
                              header=True)

    return


def recent_ap_f107_download(name, date_array, data_path,
                            mock_download_dir=None):
    """Download 45-day ap and F10.7 data from SWPC.

    Parameters
    ----------
    name : str
        Instrument name, expects 'f107' or 'ap'.
    date_array : array-like or pandas.DatetimeIndex
        Array-like or index of datetimes to be downloaded.
    data_path : str
        Path to data directory.
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded (default=None)

    Raises
    ------
    IOError
        If an unknown mock download directory is supplied or the desored file
        is missing.

    Note
    ----
    Note that the download path for the complementary Instrument will use
    the standard pysat data paths

    """
    pysat.logger.info(forecast_warning)

    # Get the file paths
    file_paths = {data_name: data_path if name == data_name else
                  general.get_instrument_data_path(
                      'sw_{:s}'.format(data_name), tag='45day')
                  for data_name in ['f107', 'ap']}

    # Check that the directories exist
    for data_path in file_paths.values():
        pysat.utils.files.check_and_make_path(data_path)

    # Get the file information
    raw_txt = general.get_local_or_remote_text(
        'https://services.swpc.noaa.gov/text/', mock_download_dir,
        '45-day-ap-forecast.txt')

    if raw_txt is None:
        pysat.logger.info("".join(["Data not downloaded for ",
                                   "45-day-ap-forecast.txt, data may have been",
                                   " saved to an unexpected filename."]))
    else:
        # Parse text to get the date the prediction was generated
        date_str = raw_txt.split(':Issued: ')[-1].split(' UTC')[0]
        dl_date = dt.datetime.strptime(date_str, '%Y %b %d %H%M')

        # Get to the forecast data
        raw_data = raw_txt.split('45-DAY AP FORECAST')[-1]

        # Grab Ap part
        raw_ap = raw_data.split('45-DAY F10.7 CM FLUX FORECAST')[0]
        raw_ap = raw_ap.split('\n')[1:-1]

        # Get the F107
        raw_f107 = raw_data.split('45-DAY F10.7 CM FLUX FORECAST')[-1]
        raw_f107 = raw_f107.split('\n')[1:-4]

        # Parse the data
        ap_times, ap = parse_45day_block(raw_ap)
        f107_times, f107 = parse_45day_block(raw_f107)

        # Save the data in DataFrames
        data = {'ap': pds.DataFrame(ap, index=ap_times, columns=['daily_Ap']),
                'f107': pds.DataFrame(f107, index=f107_times, columns=['f107'])}

        # Write out the data files
        for data_name in data.keys():
            file_name = '{:s}_45day_{:s}.txt'.format(
                data_name, dl_date.strftime('%Y-%m-%d'))
            data[data_name].to_csv(os.path.join(file_paths[data_name],
                                                file_name), header=True)

    return


def parse_45day_block(block_lines):
    """Parse the data blocks used in the 45-day Ap and F10.7 Flux Forecast file.

    Parameters
    ----------
    block_lines : list
        List of lines containing data in this data block

    Returns
    -------
    dates : list
        List of dates for each date/data pair in this block
    values : list
        List of values for each date/data pair in this block

    """

    # Initialize the output
    dates = list()
    values = list()

    # Cycle through each line in this block
    for line in block_lines:
        # Split the line on whitespace
        split_line = line.split()

        # Format the dates
        dates.extend([dt.datetime.strptime(tt, "%d%b%y")
                      for tt in split_line[::2]])

        # Format the data values
        values.extend([int(vv) for vv in split_line[1::2]])

    return dates, values


def list_files(name, tag, inst_id, data_path, format_str=None):
    """List local files for Kp or ap data obtained from SWPC.

    Parameters
    ----------
    name : str
        Instrument name.
    tag : str
        String specifying the database, expects 'def' (definitive) or 'now'
        (nowcast)
    inst_id : str
        Specifies the instrument identification, not used.
    data_path : str
        Path to data directory.
    format_str : str or NoneType
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)

    Returns
    -------
    files : pysat._files.Files
        A class containing the verified available files

    """

    if format_str is None:
        format_str = '_'.join([name, tag,
                               '{year:04d}-{month:02d}-{day:02d}.txt'])
    files = pysat.Files.from_os(data_path=data_path, format_str=format_str)

    # Pad list of files data to include most recent file under tomorrow
    if not files.empty:
        pds_offset = dt.timedelta(days=1)
        files.loc[files.index[-1] + pds_offset] = files.values[-1]
        files.loc[files.index[-1] + pds_offset] = files.values[-1]

    return files



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Supports Cp index values.

Properties
----------
platform
    'sw'
name
    'cp'
tag
    - 'def' Definitive Cp data from GFZ
    - 'now' Nowcast Cp data from GFZ
inst_id
    - ''

Note
----
Downloads data from ftp.gfz-potsdam.de (GFZ). These files also contain ap and Kp
data, and so the additional data files will be saved to the appropriate data
directories to avoid multiple downloads.

The historic definitive and nowcast Cp files are stored in yearly files, with
the current year being updated remotely on a regular basis.  If you are using
historic data for the current year, we recommend re-downloading it before
performing your data processing.

Examples
--------
::

    cp_ind = pysat.Instrument('sw', 'cp', tag='def')
    cp_ind.download(start=dt.datetime(2010, 1, 1))
    cp_ind.load(2010, 1)

"""

import datetime as dt
import functools
import numpy as np

import pysat

from pysatSpaceWeather.instruments import methods

# ----------------------------------------------------------------------------
# Instrument attributes

platform = 'sw'
name = 'cp'
tags = {'def': 'Definitive Cp data from GFZ',
        'now': 'Nowcast Cp data from GFZ'}
inst_ids = {'': list(tags.keys())}

# ----------------------------------------------------------------------------
# Instrument test attributes

# Set test dates
_test_dates = {'': {'def': dt.datetime(2009, 1, 1),
                    'now': dt.datetime(2020, 1, 1)}}

# ----------------------------------------------------------------------------
# Instrument methods

preprocess = methods.general.preprocess


def init(self):
    """Initialize the Instrument object with instrument specific values."""

    self.acknowledgements = methods.kp_ap.acknowledgements(self.name, self.tag)
    self.references = methods.kp_ap.references(self.name, self.tag)
    pysat.logger.info(self.acknowledgements)
    return


def clean(self):
    """Clean the Kp, not required for this index (empty function)."""

    return


# ----------------------------------------------------------------------------
# Instrument functions

download = functools.partial(methods.gfz.kp_ap_cp_download, platform, name)
list_files = functools.partial(methods.gfz.kp_ap_cp_list_files, name)


def load(fnames, tag='', inst_id=''):
    """Load Cp index files.

    Parameters
    ----------
    fnames : pandas.Series
        Series of filenames.
    tag : str
        Instrument tag. (default='')
    inst_id : str
        Instrument ID, not used. (default='')

    Returns
    -------
    data : pandas.DataFrame
        Object containing satellite data
    meta : pysat.Meta
        Object containing metadata such as column names and units

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    # Load the definitive or nowcast data. The Cp data stored in yearly
    # files, and we need to return data daily.  The daily date is
    # attached to filename.  Parse off the last date, load month of data,
    # and downselect to the desired day
    data = methods.gfz.load_def_now(fnames)

    # Initalize the meta data
    meta = pysat.Meta()
    methods.kp_ap.initialize_bartel_metadata(meta, 'Bartels_solar_rotation_num')
    methods.kp_ap.initialize_bartel_metadata(meta,
                                             'day_within_Bartels_rotation')

    meta['Cp'] = {
        meta.labels.units: '',
        meta.labels.name: 'Cp index',
        meta.labels.desc: ''.join(['the daily planetary character figure, ',
                                   'a qualitative estimate of the overall ',
                                   'level of geomagnetic activity for ',
                                   'this day determined from the sum of ',
                                   'the eight ap amplitudes']),
        meta.labels.min_val: 0.0,
        meta.labels.max_val: 2.5,
        meta.labels.fill_val: np.nan}
    meta['C9'] = {
        meta.labels.units: '',
        meta.labels.name: 'C9 index',
        meta.labels.desc: ''.join(['the contracted scale for Cp']),
        meta.labels.min_val: 0,
        meta.labels.max_val: 9,
        meta.labels.fill_val: -1}

    return data, meta



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Provides routines that support general space weather instruments."""

import importlib
import numpy as np
import os
import requests

import pysat


def preprocess(inst):
    """Preprocess the meta data by replacing the file fill values with NaN.

    Parameters
    ----------
    inst : pysat.Instrument
        pysat.Instrument object

    """

    # Replace all fill values with NaN
    for col in inst.variables:
        fill_val = inst.meta[col, inst.meta.labels.fill_val]

        # Ensure we are dealing with a float for future nan comparison
        if isinstance(fill_val, np.floating) or isinstance(fill_val, float):
            if ~np.isnan(fill_val):
                inst.data[col] = inst.data[col].replace(fill_val, np.nan)
                inst.meta[col] = {inst.meta.labels.fill_val: np.nan}

    return


def get_instrument_data_path(inst_mod_name, tag='', inst_id='', **kwargs):
    """Get the `data_path` attribute from an Instrument sub-module.

    Parameters
    ----------
    inst_mod_name : str
        pysatSpaceWeather Instrument module name
    tag : str
        String specifying the Instrument tag (default='')
    inst_id : str
        String specifying the instrument identification (default='')
    kwargs : dict
        Optional additional kwargs that may be used to initialize an Instrument

    Returns
    -------
    data_path : str
        Path where the Instrument data is stored

    """

    # Import the desired instrument module by name
    inst_mod = importlib.import_module(".".join(["pysatSpaceWeather",
                                                 "instruments", inst_mod_name]))

    # Initialize a temporary instrument to obtain pysat configuration
    temp_inst = pysat.Instrument(inst_module=inst_mod, tag=tag, inst_id=inst_id,
                                 **kwargs)

    # Save the data path for this Instrument down to the inst_id level
    data_path = temp_inst.files.data_path

    # Delete the temporary instrument
    del temp_inst

    return data_path


def get_local_or_remote_text(url, mock_download_dir, filename):
    """Retrieve text from a remote or local file.

    Parameters
    ----------
    filename : str
        Filename without any directory structure
    url : str
        Remote URL where file is located
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded (default=None)

    Returns
    -------
    raw_txt : str or NoneType
        All the text from the desired file or None if the file could not be
        retrieved

    Raises
    ------
    IOError
        If an unknown mock download directory is supplied.

    """
    if mock_download_dir is None:
        # Set the download webpage
        furl = ''.join([url, filename])
        req = requests.get(furl)

        if req.text.find('not found on this server') > 0:
            # Ensure useful information about server is passed on to user
            pysat.logger.warning('File {:} not found: {:}'.format(filename,
                                                                  url))
            raw_txt = None
        else:
            raw_txt = req.text if req.ok else None
    else:
        if not os.path.isdir(mock_download_dir):
            raise IOError('file location is not a directory: {:}'.format(
                mock_download_dir))

        furl = os.path.join(mock_download_dir, filename)

        if os.path.isfile(furl):
            with open(furl, 'r') as fpin:
                raw_txt = fpin.read()
        else:
            raw_txt = None

    return raw_txt



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Supports Dst values. Downloads data from NGDC.

Properties
----------
platform
    'sw'
name
    'dst'
tag
    - 'noaa' Historic Dst data coalated by and maintained by NOAA/NCEI
    - 'lasp' Predicted Dst from real-time ACE or DSCOVR provided by LASP
inst_id
    - ''

Note
----
Will only work until 2057.

"""

import datetime as dt
import ftplib
import numpy as np
import os
import pandas as pds
import shutil

import pysat

from pysatSpaceWeather.instruments.methods import dst as mm_dst
from pysatSpaceWeather.instruments.methods import lasp

# ----------------------------------------------------------------------------
# Instrument attributes

platform = 'sw'
name = 'dst'
tags = {'noaa': 'Historic Dst data coalated by and maintained by NOAA/NCEI',
        'lasp': 'Predicted Dst from real-time ACE or DSCOVR provided by LASP'}
inst_ids = {'': [tag for tag in tags.keys()]}

# Generate today's date to support loading predicted data sets
today = pysat.utils.time.today()
tomorrow = today + dt.timedelta(days=1)

# ----------------------------------------------------------------------------
# Instrument test attributes

_test_dates = {'': {'noaa': dt.datetime(2000, 1, 1), 'lasp': today}}

# Other tags assumed to be True
_test_download_ci = {'': {'noaa': False}}

# ----------------------------------------------------------------------------
# Instrument methods


def init(self):
    """Initialize the Instrument object with instrument specific values."""

    self.acknowledgements = mm_dst.acknowledgements(self.tag)
    self.references = mm_dst.references(self.tag)
    pysat.logger.info(self.acknowledgements)
    return


def clean(self):
    """Clean the Dst index, empty function."""

    return


# ----------------------------------------------------------------------------
# Instrument functions


def load(fnames, tag='', inst_id=''):
    """Load the Dst index files.

    Parameters
    ----------
    fnames : pandas.Series
        Series of filenames
    tag : str
        Instrument tag string. (default='')
    inst_id : str
        Instrument ID, not used. (default='')

    Returns
    -------
    data : pandas.DataFrame
        Object containing satellite data
    pysat.Meta
        Object containing metadata such as column names and units

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    all_data = []

    if tag == 'noaa':
        # NOAA Dst data is actually stored by year, but users can load by day.
        # Extract the actual dates from the input list of filenames as well as
        # the names of the actual files.
        fdates = []
        ufnames = []
        for filename in fnames:
            fdates.append(dt.datetime.strptime(filename[-10:], '%Y-%m-%d'))
            ufnames.append(filename[0:-11])

            # Get unique filenames that map to actual data
            ufnames = np.unique(ufnames).tolist()

        # Load unique files
        for fname in ufnames:
            with open(fname) as open_f:
                lines = open_f.readlines()
                idx = 0

                # Check if all lines are good
                max_lines = 0
                for line in lines:
                    if len(line) > 1:
                        max_lines += 1

                # Prep memory
                yr = np.zeros(max_lines * 24, dtype=int)
                mo = np.zeros(max_lines * 24, dtype=int)
                day = np.zeros(max_lines * 24, dtype=int)
                ut = np.zeros(max_lines * 24, dtype=int)
                dst = np.zeros(max_lines * 24, dtype=int)

                # Read data
                for line in lines:
                    if len(line) > 1:
                        temp_year = np.int64(line[14:16] + line[3:5])
                        if temp_year > 57:
                            temp_year += 1900
                        else:
                            temp_year += 2000

                        yr[idx:idx + 24] = temp_year
                        mo[idx:idx + 24] = np.int64(line[5:7])
                        day[idx:idx + 24] = np.int64(line[8:10])
                        ut[idx:idx + 24] = np.arange(24)
                        temp = line.strip()[20:-4]
                        temp2 = [temp[4 * i:4 * (i + 1)] for i in np.arange(24)]
                        dst[idx:idx + 24] = temp2
                        idx += 24

                # Prep datetime index for the data and create DataFrame
                start = dt.datetime(yr[0], mo[0], day[0], ut[0])
                stop = dt.datetime(yr[-1], mo[-1], day[-1], ut[-1])
                dates = pds.date_range(start, stop, freq='H')
                new_data = pds.DataFrame(dst, index=dates, columns=['dst'])

                # Add to all data loaded for filenames
                all_data.append(new_data)

        # Combine data together
        data = pds.concat(all_data, sort=True, axis=0)

        # Pull out requested days
        data = data.iloc[data.index >= fdates[0], :]
        data = data.iloc[data.index < fdates[-1] + pds.DateOffset(days=1), :]
    else:
        data = pysat.instruments.methods.general.load_csv_data(
            fnames, read_csv_kwargs={'index_col': 0, 'parse_dates': True})

    # Create metadata
    meta = pysat.Meta()
    meta['dst'] = {meta.labels.units: 'nT',
                   meta.labels.name: 'Dst',
                   meta.labels.notes: tags[tag],
                   meta.labels.desc: 'Disturbance storm-time index',
                   meta.labels.fill_val: np.nan,
                   meta.labels.min_val: -np.inf,
                   meta.labels.max_val: np.inf}

    return data, meta


def list_files(tag='', inst_id='', data_path='', format_str=None):
    """List local data files for Dst data.

    Parameters
    ----------
    tag : str
        Instrument tag, accepts any value from `tags`. (default='')
    inst_id : str
        Instrument ID, not used. (default='')
    data_path : str
        Path to data directory. (default='')
    format_str : str or NoneType
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)

    Returns
    -------
    files : pysat.Files
        A class containing the verified available files

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """
    # Get the format string, if not supplied by the user
    if format_str is None:
        if tag == 'noaa':
            format_str = 'dst{year:4d}.txt'
        else:
            format_str = ''.join(['sw_dst_', tag, '_{year:4d}-{month:2d}-',
                                  '{day:2d}.txt'])

    # Get the desired files
    files = pysat.Files.from_os(data_path=data_path, format_str=format_str)

    if tag == 'noaa':
        # NOAA files yearly, so we need to add daily dates to the yearly
        # filenames. The load routine will load a month of data and use
        # the appended date to select out appropriate data.
        if not files.empty:
            files.loc[files.index[-1] + pds.DateOffset(years=1)
                      - pds.DateOffset(days=1)] = files.iloc[-1]
            files = files.asfreq('D', 'pad')
            files = files + '_' + files.index.strftime('%Y-%m-%d')

    return files


def download(date_array, tag, inst_id, data_path, mock_download_dir=None):
    """Download the Dst index data from the appropriate repository.

    Parameters
    ----------
    date_array : array-like or pandas.DatetimeIndex
        Array-like or index of datetimes for which files will be downloaded.
    tag : str
        Instrument tag, used to determine download location.
    inst_id : str
        Instrument ID, not used.
    data_path : str
        Path to data directory.
        If not None, will process any files with the correct name and date
        as if they were downloaded.
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded. (default=None)

    Raises
    ------
    IOError
        If an unknown mock download directory is supplied.

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """
    # If a mock download directory was supplied, test to see it exists
    if mock_download_dir is not None:
        if not os.path.isdir(mock_download_dir):
            raise IOError('file location is not a directory: {:}'.format(
                mock_download_dir))

    if tag == 'noaa':
        if mock_download_dir is None:
            # Connect to host, default port
            ftp = ftplib.FTP('ftp.ngdc.noaa.gov')

            # User anonymous, passwd anonymous@
            ftp.login()
            ftp.cwd('/STP/GEOMAGNETIC_DATA/INDICES/DST')

        # Data stored by year. Only download for unique set of input years.
        years = np.array([date.year for date in date_array])
        years = np.unique(years)
        for year in years:
            fname_root = 'dst{year:04d}.txt'
            fname = fname_root.format(year=year)
            saved_fname = os.path.join(data_path, fname)
            if mock_download_dir is None:
                try:
                    pysat.logger.info('Downloading file for {year:04d}'.format(
                        year=year))
                    with open(saved_fname, 'wb') as fp:
                        ftp.retrbinary('RETR ' + fname, fp.write)
                except ftplib.error_perm as exception:
                    if str(exception.args[0]).split(" ", 1)[0] != '550':
                        raise exception
                    else:
                        # File not present
                        os.remove(saved_fname)
                        pysat.logger.info(
                            'File not available for {:04d}'.format(year))
            else:
                # Get the local file, if it exists
                down_fname = os.path.join(mock_download_dir, fname)
                if os.path.isfile(down_fname):
                    shutil.copyfile(down_fname, saved_fname)
                else:
                    pysat.logger.info("".join(["Data not downloaded for ",
                                               down_fname, ", data may have ",
                                               "been saved to an unexpected ",
                                               "filename."]))

        if mock_download_dir is None:
            ftp.close()
    elif tag == 'lasp':
        lasp.prediction_downloads(name, tag, data_path, mock_download_dir)

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Supports polar cap indexes.

Properties
----------
platform
    'sw'
name
    'polarcap'
tag
    - 'prediction' Predictions from SWPC for the next 3 days
inst_id
    - ''

Note
----
Downloads data from SWPC. These files also contain other data indices, and so
the additional data files will be saved to the appropriate data directories to
avoid multiple downloads.

The forecast/prediction data is stored by generation date, where each file
contains the forecast for the next three days. Forecast data downloads are only
supported for the current day. When loading forecast data, the date specified
with the load command is the date the forecast was generated. The data loaded
will span three days. To always ensure you are loading the most recent data,
load the data with tomorrow's date.

Examples
--------
::

    pc = pysat.Instrument('sw', 'polarcap', tag='prediction')
    pc.download()
    pc.load(date=pc.tomorrow())


Warnings
--------
The 'prediction' tag loads polar cap absoprtion predictions for a specific
period of time. Loading multiple files, loading multiple days, the data padding
feature, and multi_file_day feature available from the pyast.Instrument object
is not appropriate for these tags data.

"""

import datetime as dt
import functools
import numpy as np
import pandas as pds

import pysat

from pysatSpaceWeather.instruments import methods

# ----------------------------------------------------------------------------
# Instrument attributes

platform = 'sw'
name = 'polarcap'
tags = {'prediction': 'SWPC Predictions for the next three days'}
inst_ids = {'': list(tags.keys())}

# Generate todays date to support loading forecast data
now = dt.datetime.utcnow()
today = dt.datetime(now.year, now.month, now.day)
tomorrow = today + dt.timedelta(days=1)

# ----------------------------------------------------------------------------
# Instrument test attributes

# Set test dates
_test_dates = {'': {'prediction': tomorrow}}

# ----------------------------------------------------------------------------
# Instrument methods

preprocess = methods.general.preprocess


def init(self):
    """Initialize the Instrument object with instrument specific values."""

    self.acknowledgements = methods.swpc.ackn
    self.references = "".join(["Check SWPC for appropriate references: ",
                               "https://www.swpc.noaa.gov/phenomena"])
    pysat.logger.info(self.acknowledgements)
    return


def clean(self):
    """Clean the polar cap indexes, not required (empty function)."""

    return


# ----------------------------------------------------------------------------
# Instrument functions

list_files = functools.partial(methods.swpc.list_files, name)


def load(fnames, tag='', inst_id=''):
    """Load storm probability files.

    Parameters
    ----------
    fnames : pandas.Series
        Series of filenames
    tag : str
        Instrument tag (default='')
    inst_id : str
        Instrument ID, not used. (default='')

    Returns
    -------
    data : pandas.DataFrame
        Object containing satellite data
    meta : pysat.Meta
        Object containing metadata such as column names and units

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    Warnings
    --------
    tag '' has been deprecated, will be removed in version 0.2.0+

    """

    # Load the data
    all_data = []
    for fname in fnames:
        result = pds.read_csv(fname, index_col=0, parse_dates=True)
        all_data.append(result)

    data = pds.concat(all_data)

    # Initialize the metadata
    meta = pysat.Meta()
    for lkey in ['fill_val', 'min_val', 'max_val']:
        meta.labels.label_type[lkey] = (float, int, np.float64, np.int64, str)

    for dkey in data.columns:
        meta[dkey] = {meta.labels.units: '',
                      meta.labels.name: dkey,
                      meta.labels.desc: dkey.replace('_', ' '),
                      meta.labels.min_val: 'green',
                      meta.labels.max_val: 'red',
                      meta.labels.fill_val: ''}

    return data, meta


def download(date_array, tag, inst_id, data_path, mock_download_dir=None):
    """Download the polar cap indices from the appropriate repository.

    Parameters
    ----------
    date_array : array-like or pandas.DatetimeIndex
        Array-like or index of datetimes to be downloaded.
    tag : str
        Denotes type of file to load.
    inst_id : str
        Specifies the instrument identification, not used.
    data_path : str
        Path to data directory.
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded. (default=None)

    Raises
    ------
    IOError
        If an unknown mock download directory is supplied.

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    Warnings
    --------
    Only able to download current recent data, not archived forecasts.

    """

    if tag == 'prediction':
        methods.swpc.solar_geomag_predictions_download(
            name, date_array, data_path, mock_download_dir)

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Supports solar magnetic field values.

Properties
----------
platform
    'sw'
name
    'sbfield'
tag
    - 'prelim' Preliminary SWPC daily solar indices
    - 'daily' Daily SWPC solar indices (contains last 30 days)

Examples
--------
Download and load some preliminary solar magnetic field data.
::

    smf = pysat.Instrument('sw', 'sbfield', tag='prelim')
    smf.download(start=dt.datetime(2001, 1, 1))
    smf.load(date=dt.datetime(2001, 1, 1))

"""

import datetime as dt
import numpy as np
import os
import pandas as pds

import pysat

from pysatSpaceWeather.instruments import methods

# ----------------------------------------------------------------------------
# Instrument attributes

platform = 'sw'
name = 'sbfield'
tags = {'prelim': 'Preliminary SWPC daily solar indices',
        'daily': 'Daily SWPC solar indices (contains last 30 days)'}

# Dict keyed by inst_id that lists supported tags for each inst_id
inst_ids = {'': [tag for tag in tags.keys()]}

# Dict keyed by inst_id that lists supported tags and a good day of test data
# generate todays date to support loading forecast data
now = dt.datetime.utcnow()
today = dt.datetime(now.year, now.month, now.day)
tomorrow = today + pds.DateOffset(days=1)

# ----------------------------------------------------------------------------
# Instrument test attributes

_test_dates = {'': {'prelim': dt.datetime(2009, 1, 1),
                    'daily': tomorrow}}

# Other tags assumed to be True
_test_download_ci = {'': {'prelim': False}}

# ----------------------------------------------------------------------------
# Instrument methods

preprocess = methods.general.preprocess


def init(self):
    """Initialize the Instrument object with instrument specific values."""

    # Set the required Instrument attributes
    self.acknowledgements = methods.swpc.ackn
    self.references = "".join(["Scherrer, P.H., Wilcox, J.M., Svalgaard, L. ",
                               "et al. The mean magnetic field of the Sun: ",
                               "Observations at Stanford. Sol Phys 54, 353–361",
                               " (1977). https://doi.org/10.1007/BF00159925"])
    pysat.logger.info(self.acknowledgements)

    return


def clean(self):
    """Clean the data, empty function as this is not necessary."""

    return


# ----------------------------------------------------------------------------
# Instrument functions


def load(fnames, tag='', inst_id=''):
    """Load solar magnetic field files.

    Parameters
    ----------
    fnames : pandas.Series
        Series of filenames.
    tag : str
        Instrument tag. (default='')
    inst_id : str
        Instrument ID, not used. (default='')

    Returns
    -------
    data : pandas.DataFrame
        Object containing satellite data.
    meta : pysat.Meta
        Object containing metadata such as column names and units.

    See Also
    --------
    pysat.instruments.methods.general.load_csv_data

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    # Get the desired file dates and file names from the daily indexed list
    file_dates = list()
    if tag == 'prelim':
        unique_files = list()
        for fname in fnames:
            file_dates.append(dt.datetime.strptime(fname[-10:], '%Y-%m-%d'))
            if fname[0:-11] not in unique_files:
                unique_files.append(fname[0:-11])
        fnames = unique_files

    # Load the CSV data files
    data = pysat.instruments.methods.general.load_csv_data(
        fnames, read_csv_kwargs={"index_col": 0, "parse_dates": True})

    # If there is a date range, downselect here
    if len(file_dates) > 0:
        idx, = np.where((data.index >= min(file_dates))
                        & (data.index < max(file_dates) + dt.timedelta(days=1)))
        data = data.iloc[idx, :]

    # Initialize the metadata
    meta = pysat.Meta()
    meta['smf'] = {meta.labels.units: 'G',
                   meta.labels.name: 'Solar Mean Field',
                   meta.labels.notes: '',
                   meta.labels.desc: 'Standford Solar Mean Field',
                   meta.labels.fill_val: -999,
                   meta.labels.min_val: 0,
                   meta.labels.max_val: np.inf}

    return data, meta


def list_files(tag='', inst_id='', data_path='', format_str=None):
    """List local solar magnetic field data files.

    Parameters
    ----------
    tag : str
        Instrument tag, accepts any value from `tags`. (default='')
    inst_id : str
        Instrument ID, not used. (default='')
    data_path : str
        Path to data directory. (default='')
    format_str : str or NoneType
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)

    Returns
    -------
    out_files : pysat._files.Files
        A class containing the verified available files

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    if tag == 'prelim':
        # Files are by year (and quarter)
        if format_str is None:
            format_str = ''.join(['sbfield_prelim_{year:04d}_{month:02d}',
                                  '_v{version:01d}.txt'])
        out_files = pysat.Files.from_os(data_path=data_path,
                                        format_str=format_str)

        if not out_files.empty:
            # Set each file's valid length at a 1-day resolution
            orig_files = out_files.sort_index().copy()
            new_files = list()

            for orig in orig_files.items():
                # Version determines each file's valid length
                version = int(orig[1].split("_v")[1][0])
                doff = pds.DateOffset(years=1) if version == 2 \
                    else pds.DateOffset(months=3)
                istart = orig[0]
                iend = istart + doff - pds.DateOffset(days=1)

                # Ensure the end time does not extend past the number of
                # possible days included based on the file's download time
                fname = os.path.join(data_path, orig[1])
                dend = dt.datetime.utcfromtimestamp(os.path.getctime(fname))
                dend = dend - pds.DateOffset(days=1)
                if dend < iend:
                    iend = dend

                # Pad the original file index
                out_files.loc[iend] = orig[1]
                out_files = out_files.sort_index()

                # Save the files at a daily cadence over the desired period
                new_files.append(out_files.loc[istart:
                                               iend].asfreq('D', 'pad'))
            # Add the newly indexed files to the file output
            out_files = pds.concat(new_files, sort=True)
            out_files = out_files.dropna()
            out_files = out_files.sort_index()
            out_files = out_files + '_' + out_files.index.strftime('%Y-%m-%d')

    elif tag == 'daily':
        out_files = methods.swpc.list_files(name, tag, inst_id, data_path,
                                            format_str=format_str)

    return out_files


def download(date_array, tag, inst_id, data_path, update_files=False,
             mock_download_dir=None):
    """Download solar magnetic field data from the appropriate repository.

    Parameters
    ----------
    date_array : array-like
        Sequence of dates for which files will be downloaded.
    tag : str
        Denotes type of file to load.
    inst_id : str
        Specifies the satellite ID for a constellation.
    data_path : str
        Path to data directory.
    update_files : bool
        Re-download data for files that already exist if True (default=False)
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded. (default=None)

    Raises
    ------
    IOError
        If a problem is encountered connecting to the gateway or retrieving
        data from the remote or local repository.

    Warnings
    --------
    Only able to download current forecast data, not archived forecasts.

    """
    if tag == 'prelim':
        # Get the local files, to ensure that the version 1 files are
        # downloaded again if more data has been added
        local_files = list_files(tag, inst_id, data_path)

        methods.swpc.old_indices_dsd_download(name, date_array, data_path,
                                              local_files, today,
                                              mock_download_dir)

    elif tag == 'daily':
        methods.swpc.daily_dsd_download(name, today, data_path,
                                        mock_download_dir)

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Provides general routines for the ACE space weather instruments."""

import datetime as dt
import numpy as np
import os
import pandas as pds

import pysat

from pysatSpaceWeather.instruments.methods import general

logger = pysat.logger
clean_warn = {'dusty': [('logger', 'WARN',
                         "unused clean level 'dusty', reverting to 'clean'",
                         'clean')]}


def acknowledgements():
    """Define the acknowledgements for the specified ACE instrument.

    Returns
    -------
    ackn : str
        Acknowledgements for the ACE instrument

    """

    ackn = ''.join(['NOAA provided funds for the modification of ',
                    ' the ACE transmitter to enable the broadcast of',
                    ' the real-time data and also funds to the ',
                    'instrument teams to provide the algorithms for ',
                    'processing the real-time data.'])

    return ackn


def references(name):
    """Define the references for the specified ACE instrument.

    Parameters
    ----------
    name : str
        Instrument name of the ACE instrument

    Returns
    -------
    ref : str
        Reference for the ACE instrument paper

    """

    refs = {'mag': "".join(["'The ACE Magnetic Field Experiment', ",
                            "C. W. Smith, M. H. Acuna, L. F. Burlaga, ",
                            "J. L'Heureux, N. F. Ness and J. Scheifele, ",
                            "Space Sci. Rev., 86, 613-632 (1998)."]),
            'epam': ''.join(['Gold, R. E., S. M. Krimigis, S. E. Hawkins, ',
                             'D. K. Haggerty, D. A. Lohr, E. Fiore, ',
                             'T. P. Armstrong, G. Holland, L. J. Lanzerotti,',
                             ' Electron, Proton and Alpha Monitor on the ',
                             'Advanced Composition Explorer Spacecraft, ',
                             'Space Sci. Rev., 86, 541, 1998.']),
            'swepam': ''.join(['McComas, D., Bame, S., Barker, P. et al. ',
                               'Solar Wind Electron Proton Alpha Monitor ',
                               '(SWEPAM) for the Advanced Composition ',
                               'Explorer. Space Sci. Rev., 86, 563–612',
                               ' (1998). ',
                               'https://doi.org/10.1023/A:1005040232597']),
            'sis': ''.join(['Stone, E., Cohen, C., Cook, W. et al. The ',
                            'Solar Isotope Spectrometer for the Advanced ',
                            'Composition Explorer. Space Sci. Rev., 86, ',
                            '357–408 (1998). ',
                            'https://doi.org/10.1023/A:1005027929871'])}

    if name not in refs.keys():
        raise KeyError('unknown ACE instrument, accepts {:}'.format(
            refs.keys()))

    return refs[name]


def clean(inst):
    """Clean the common aspects of the ACE space weather data.

    Parameters
    ----------
    inst : pysat.Instrument
        ACE pysat.Instrument object

    Returns
    -------
    max_status : int
        Maximum allowed status

    Note
    ----
    pysat Instrument is modified in place

    """

    if inst.platform != "ace":
        raise AttributeError("Can't apply ACE cleaning to platform {:}".format(
            inst.platform))

    # Set the clean level
    if inst.clean_level == 'dusty':
        logger.warning("unused clean level 'dusty', reverting to 'clean'")
        inst.clean_level = 'clean'

    max_status = 9
    if inst.clean_level == "clean":
        max_status = 0
    elif inst.clean_level == "dirty":
        max_status = 8

    return max_status


def list_files(name, tag='', inst_id='', data_path='', format_str=None):
    """List the local ACE data files.

    Parameters
    ----------
    name : str
        ACE Instrument name.
    tag : str
        ACE Instrument tag. (default='')
    inst_id : str
        Specifies the ACE instrument ID. (default='')
    data_path : str
        Path to data directory. (default='')
    format_str : str or NoneType
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)

    Returns
    -------
    files : pysat.Files
        A class containing the verified available files

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """
    if format_str is None:
        format_str = '_'.join(["ace", name, tag,
                               '{year:04d}-{month:02d}-{day:02d}.txt'])

    files = pysat.Files.from_os(data_path=data_path, format_str=format_str)

    return files


def download(date_array, name, tag='', inst_id='', data_path='', now=None,
             mock_download_dir=None):
    """Download the requested ACE Space Weather data.

    Parameters
    ----------
    date_array : array-like
        Array of datetime values
    name : str
        ACE Instrument name.
    tag : str
        ACE Instrument tag. (default='')
    inst_id : str
        ACE instrument ID. (default='')
    data_path : str
        Path to data directory. (default='')
    now : dt.datetime or NoneType
        Current universal time, if None this is determined for each
        download. (default=None)
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded (default=None)

    Raises
    ------
    IOError
        If an unknown mock download directory is supplied or the file format
        changes.

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    Warnings
    --------
    - Only able to download current real-time data
    - File requested not available on server

    """
    # Ensure now is up-to-date, if desired
    if now is None:
        now = dt.datetime.utcnow()

    # Define the file information for each data type and check the
    # date range
    if tag == 'realtime':
        file_fmt = "{:s}-{:s}.txt".format("ace", "magnetometer"
                                          if name == "mag" else name)

        if any([len(date_array) > 1, date_array[0].year != now.year,
                date_array[0].month != now.month,
                date_array[0].day != now.day]):
            logger.warning(''.join(['real-time data only available for current',
                                    ' day, data in this file will have the ',
                                    'wrong date.']))
    else:
        data_rate = 1 if name in ['mag', 'swepam'] else 5
        file_fmt = '_'.join(["%Y%m%d", "ace", name,
                             '{:d}m.txt'.format(data_rate)])

    url = {'realtime': 'https://services.swpc.noaa.gov/text/',
           'historic': 'https://sohoftp.nascom.nasa.gov/sdb/ace/daily/'}

    data_cols = {'mag': ['jd', 'sec', 'status', 'bx_gsm', 'by_gsm',
                         'bz_gsm', 'bt_gsm', 'lat_gsm', 'lon_gsm'],
                 "swepam": ['jd', 'sec', 'status', 'sw_proton_dens',
                            'sw_bulk_speed', 'sw_ion_temp'],
                 "epam": ['jd', 'sec', 'status_e', 'eflux_38-53',
                          'eflux_175-315', 'status_p', 'pflux_47-68',
                          'pflux_115-195', 'pflux_310-580',
                          'pflux_795-1193', 'pflux_1060-1900', 'anis_ind'],
                 'sis': ['jd', 'sec', 'status_10', 'int_pflux_10MeV',
                         'status_30', 'int_pflux_30MeV']}

    # Cycle through all the dates
    for dl_date in date_array:
        # Get the file text from the remote or local destination
        raw_data = general.get_local_or_remote_text(url[tag], mock_download_dir,
                                                    dl_date.strftime(file_fmt))

        if raw_data is None:
            pysat.logger.info("".join(["Data not downloaded for ",
                                       dl_date.strftime(file_fmt), ", date may",
                                       " be out of range for the database or ",
                                       "data may have been saved to an ",
                                       "unexpected filename. Check URL: ",
                                       url[tag], ", or directory: ",
                                       repr(mock_download_dir)]))
        else:
            # Split the file at the last header line and the new line markers
            raw_data = raw_data.split('#-----------------')[-1]
            raw_data = raw_data.split('\n')[1:]  # Remove the last header line

            # Parse the file, treating the 4 time columns separately
            data_dict = {col: list() for col in data_cols[name]}
            times = list()
            nsplit = len(data_cols[name]) + 4
            for raw_line in raw_data:
                split_line = raw_line.split()
                if len(split_line) == nsplit:
                    times.append(dt.datetime.strptime(
                        ' '.join(split_line[:4]), '%Y %m %d %H%M'))
                    for i, col in enumerate(data_cols[name]):
                        # Convert to a number and save
                        #
                        # Output is saved as a float, so don't bother to
                        # differentiate between int and float
                        data_dict[col].append(float(split_line[4 + i]))
                else:
                    if len(split_line) > 0:
                        raise IOError(''.join([
                            'unexpected line encoutered in ', url[tag], "/",
                            dl_date.strftime(file_fmt), ":\n", raw_line]))

            # Put data into nicer DataFrame
            data = pds.DataFrame(data_dict, index=times)

            # Write out as a file
            data_file = '{:s}.txt'.format(
                '_'.join(["ace", name, tag, dl_date.strftime('%Y-%m-%d')]))
            data.to_csv(os.path.join(data_path, data_file), header=True)

    return


def common_metadata():
    """Define the common metadata information for all ACE instruments.

    Returns
    -------
    meta : pysat.Meta
        pysat Meta class with 'jd' and 'sec' initiated
    status_desc : str
        Description of the status flags

    """
    # Initialize the metadata
    meta = pysat.Meta()

    # Define the Julian day
    meta['jd'] = {meta.labels.units: 'days',
                  meta.labels.name: 'MJD',
                  meta.labels.notes: '',
                  meta.labels.desc: 'Modified Julian Day',
                  meta.labels.fill_val: np.nan,
                  meta.labels.min_val: -np.inf,
                  meta.labels.max_val: np.inf}

    # Define the seconds of day
    meta['sec'] = {meta.labels.units: 's',
                   meta.labels.name: 'Sec of Day',
                   meta.labels.notes: '',
                   meta.labels.desc: 'Seconds of Julian Day',
                   meta.labels.fill_val: np.nan,
                   meta.labels.min_val: -np.inf,
                   meta.labels.max_val: np.inf}

    # Provide information about the status flags
    status_desc = '0 = nominal data, 1 to 8 = bad data record, 9 = no data'

    return meta, status_desc


def ace_swepam_hourly_omni_norm(as_inst, speed_key='sw_bulk_speed',
                                dens_key='sw_proton_dens',
                                temp_key='sw_ion_temp'):
    """Normalize ACE SWEPAM variables as described in the OMNI processing _[1].

    Parameters
    ----------
    as_inst : pysat.Instrument
        pysat Instrument object with ACE SWEPAM data.
    speed_key : str
        Data key for bulk solar wind speed data in km/s
        (default='sw_bulk_speed')
    dens_key : str
        Data key for solar wind proton density data in P/cm^3
        (default='sw_proton_dens')
    temp_key : str
        Data key for solar wind ion temperature data in K
        (default='sw_ion_temp')

    References
    ----------
    [1] https://omniweb.gsfc.nasa.gov/html/omni_min_data.html

    """

    # Check the input to make sure all the necessary data variables are present
    for var in [speed_key, dens_key, temp_key]:
        if var not in as_inst.variables:
            raise ValueError('instrument missing variable: {:}'.format(var))

    # Let yt be the fractional years since 1998.0
    yt = np.array([pysat.utils.time.datetime_to_dec_year(itime) - 1998.0
                   for itime in as_inst.index])

    # The normalization depends on the year
    yt_dens = (yt >= 2019.0) & (yt <= 2021.0)
    yt_temp = (yt >= 2019.0) & (yt <= 2020.0)

    # Get the masks for the different velocity limits
    ilow = as_inst[speed_key] < 395
    imid = (as_inst[speed_key] >= 395) & (as_inst[speed_key] <= 405)
    ihigh = as_inst[speed_key] > 405

    # Calculate the normalized plasma density
    norm_n = np.array(as_inst[dens_key])
    norm_n[ilow] *= (0.925 + 0.0039 * yt[ilow])
    norm_n[imid] *= (74.02 - 0.164 * as_inst[speed_key][imid]
                     + 0.0171 * as_inst[speed_key][imid] * yt[imid]
                     - 6.72 * yt[imid]) / 10.0
    norm_n[ihigh] *= (0.761 + 0.0210 * yt[ihigh])

    # Overwrite the calculation for the year where velocity isn't important
    norm_n[yt_dens] = np.power(10.0, -0.010 + 1.006
                               * np.log10(as_inst[dens_key][yt_dens]))

    # Normalize the temperature
    norm_t = np.array(as_inst[temp_key])
    norm_t[~yt_temp] = np.power(10.0,
                                -0.069 + 1.024 * np.log10(norm_t[~yt_temp]))
    norm_t[yt_temp] = np.power(10.0,
                               0.266 + 0.947 * np.log10(norm_t[yt_temp]))

    # Update the instrument data
    as_inst['sw_proton_dens_norm'] = pds.Series(norm_n, index=as_inst.index)
    as_inst['sw_ion_temp_norm'] = pds.Series(norm_t, index=as_inst.index)

    # Add meta data
    for dkey in [dens_key, temp_key]:
        nkey = '{:s}_norm'.format(dkey)
        meta_dict = {}

        for mkey in as_inst.meta[dkey].keys():
            if mkey == as_inst.meta.labels.notes:
                meta_dict[mkey] = ''.join([
                    'Normalized for hourly OMNI as described in ',
                    'https://omniweb.gsfc.nasa.gov/html/omni_min_data.html'])
            elif mkey != "children":
                meta_dict[mkey] = as_inst.meta[dkey, mkey]

        as_inst.meta[nkey] = meta_dict

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Provides support functions for the LASP LISIRD data base."""

import datetime as dt
import json
import numpy as np
import os
import pandas as pds
import requests

import pysat

ackn = "".join(["LASP Interactive Solar Irradiance Data Center provides ",
                "access to many solar datasets generated by researchers at ",
                "LASP and other institutions."])


def references(platform, name, tag, inst_id):
    """Provide references for different Instrument data products.

    Parameters
    ----------
    platform : str
        Instrument platform
    name : str
        Instrument name
    tag : str
        Instrument tag
    inst_id : str
        Instrument ID

    Returns
    -------
    refs : str
        String of references

    """

    refs = {'sw': {'mgii': {
        'composite': {
            '': ''.join(["Viereck, R. A., Floyd, L. E., Crane, P. C., Woods, ",
                         "T. N., Knapp, B. G., Rottman, G., Weber, M., Puga,",
                         " L. C., and DeLand, M. T. (2004), A composite Mg ",
                         "II index spanning from 1978 to 2003, Space Weather",
                         ", 2, S10005, doi:10.1029/2004SW000084."])},
        'sorce': {
            '': "\n".join([
                "".join(["Snow, M, William E. McClintock, Thomas N. Woods, ",
                         "Oran R. White, Jerald W. Harder, and Gary Rottman ",
                         "(2005). The Mg II Index from SORCE, Solar Phys., ",
                         "230, 1, 325-344."]),
                "".join(["Heath, D. and Schlesinger, B. (1986). The Mg 280-nm ",
                         "doublet as a monitor of changes in solar ",
                         "ultraviolet irradiance, JGR, 91, 8672-8682."])])}}}}

    return refs[platform][name][tag][inst_id]


def build_lisird_url(lisird_data_name, start, stop):
    """Build a LASP LISIRD direct download URL.

    Parameters
    ----------
    lisird_data_name : str
        Name of the data set on the LISARD server
    start : dt.datetime
        Start time
    stop : dt.datetime
        Stop time

    Returns
    -------
    url : str
        URL that will download the desired data

    """

    # Define the formatting for the start and stop times
    tfmt = "%Y-%m-%dT%H:%M:%S.000Z"

    url = "".join(["https://lasp.colorado.edu/lisird/latis/dap/",
                   lisird_data_name, ".json?&time>=", start.strftime(tfmt),
                   '&time<=', stop.strftime(tfmt),
                   "&format_time(yyyy-MM-dd'T'HH:mm:ss.SSS)"])

    return url


def download(date_array, data_path, local_file_prefix, local_date_fmt,
             lisird_data_name, freq, update_files=False, fill_vals=None,
             mock_download_dir=None):
    """Download routine for LISIRD data.

    Parameters
    ----------
    date_array : array-like
        Sequence of dates for which files will be downloaded.
    data_path : str
        Path to data directory.
    local_file_prefix : str
        Prefix for local files, e.g., 'tag_' or 'tag_monthly_'
    local_date_fmt : str
        String format for the local filename, e.g., '%Y-%m-%d' or '%Y-%m'
    lisird_data_name : str
        Name of the data set on the LISARD server
    freq : pds.DateOffset or dt.timedelta
        Offset to add to the start date to ensure all data is downloaded
        (inclusive)
    update_files : bool
        Re-download data for files that already exist if True (default=False)
    fill_vals : dict or NoneType
        Dict of fill values to replace with NaN by variable name or None to
        leave alone (default=None)
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date (following the local
        file prefix and date format) as if they were downloaded (default=None)

    Raises
    ------
    IOError
        If there is a gateway timeout when downloading data or an unknown mock
        download directory is supplied.
    KeyError
        If the `fill_vals` input does not match the downloaded data.

    """
    # If a mock download directory was supplied, test to see it exists
    if mock_download_dir is not None:
        if not os.path.isdir(mock_download_dir):
            raise IOError('file location is not a directory: {:}'.format(
                mock_download_dir))

    # Initialize the fill_vals dict, if necessary
    if fill_vals is None:
        fill_vals = {}

    # Cycle through all the dates
    for dl_date in date_array:
        # Build the local filename
        fname = ''.join([local_file_prefix, dl_date.strftime(local_date_fmt),
                         '.txt'])
        local_file = os.path.join(data_path, fname)

        # Determine if the download should occur
        if update_files or not os.path.isfile(local_file):
            if mock_download_dir is None:
                # Get the URL for the desired data
                url = build_lisird_url(lisird_data_name, dl_date,
                                       dl_date + freq)

                # The data is returned as a JSON file
                req = requests.get(url)

                # Process the JSON file
                if req.text.find('Gateway Timeout') >= 0:
                    raise IOError(''.join(['Gateway timeout when requesting ',
                                           'file using command: ', url]))

                # Load the dict if text was retrieved
                json_dict = json.loads(req.text) if req.ok else {'': {}}
            else:
                # Get the local repository filename
                url = os.path.join(mock_download_dir, fname)
                if os.path.isfile(url):
                    with open(url, 'r') as fpin:
                        raw_txt = fpin.read()

                    json_dict = json.loads(raw_txt)
                else:
                    json_dict = {'': {}}

            if lisird_data_name in json_dict.keys():
                raw_dict = json_dict[lisird_data_name]
                data = pds.DataFrame.from_dict(raw_dict['samples'])
                if data.empty:
                    pysat.logger.warning("no data for {:}".format(dl_date))
                else:
                    # The URL specifies the time format, so break it down
                    frac_sec = [int(tval.split('.')[-1])
                                for tval in data['time']]
                    times = [dt.datetime.strptime(tval.split('.')[0],
                                                  '%Y-%m-%dT%H:%M:%S')
                             + dt.timedelta(microseconds=frac_sec[i] * 6)
                             for i, tval in enumerate(data.pop('time'))]
                    data.index = times

                    # Replace fill value with NaNs
                    for var in fill_vals.keys():
                        if var in data.columns:
                            idx, = np.where(data[var] == fill_vals[var])
                            data.iloc[idx, :] = np.nan
                        else:
                            raise KeyError(''.join(['unknown fill value ',
                                                    'variable name supplied: ',
                                                    var]))

                    # Create a local CSV file
                    data.to_csv(local_file, header=True)
            else:
                if len(json_dict.keys()) == 1 and '' in json_dict.keys():
                    pysat.logger.info("".join(["Data not downloaded for ",
                                               dl_date.strftime("%d %b %Y"),
                                               ", date may be out of range ",
                                               "for the database or data may ",
                                               "have been saved to an ",
                                               "unexpected filename: ", url]))
                else:
                    raise IOError(''.join(['Returned unexpectedly formatted ',
                                           'data using command: ', url]))
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Supports F10.7 index values.

Properties
----------
platform
    'sw'
name
    'f107'
tag
    - 'historic' LASP F10.7 data (downloads by month, loads by day)
    - 'prelim' Preliminary SWPC daily solar indices
    - 'now' A mix of nowcast and definitive values from GFZ
    - 'daily' Daily SWPC solar indices (contains last 30 days)
    - 'forecast' Grab forecast data from SWPC (next 3 days)
    - '45day' 45-Day Forecast data from the Air Force
inst_id
    - '' No distinction, may include observed, adjusted, or both
    - 'obs' Observed F10.7
    - 'adj' Adjusted F10.7

Examples
--------
Download and load all of the historic F10.7 data.  Note that it will not
stop on the current date, but a point in the past when post-processing has
been successfully completed.
::

    f107 = pysat.Instrument('sw', 'f107', tag='historic')
    f107.download(start=f107.lasp_stime, stop=f107.today())
    f107.load(date=f107.lasp_stime, end_date=f107.today())


Note
----
The forecast data is stored by generation date, where each file contains the
forecast for the next three days. Forecast data downloads are only supported
for the current day. When loading forecast data, the date specified with the
load command is the date the forecast was generated. The data loaded will span
three days. To always ensure you are loading the most recent data, load
the data with tomorrow's date.
::

    f107 = pysat.Instrument('sw', 'f107', tag='forecast')
    f107.download()
    f107.load(date=f107.tomorrow())


Warnings
--------
The 'forecast' F10.7 data loads three days at a time. Loading multiple files,
loading multiple days, the data padding feature, and multi_file_day feature
available from the pyast.Instrument object is not appropriate for 'forecast'
data.

Like 'forecast', the '45day' forecast loads a specific period of time (45 days)
and subsequent files contain overlapping data.  Thus, loading multiple files,
loading multiple days, the data padding feature, and multi_file_day feature
available from the pyast.Instrument object is not appropriate for '45day' data.

"""

import datetime as dt
import numpy as np
import os
import pandas as pds

import pysat

from pysatSpaceWeather.instruments import methods

# ----------------------------------------------------------------------------
# Instrument attributes

platform = 'sw'
name = 'f107'
tags = {'historic': 'Daily LASP value of F10.7',
        'prelim': 'Preliminary SWPC daily solar indices',
        'now': 'Nowcast and definitive data from GFZ',
        'daily': 'Daily SWPC solar indices (contains last 30 days)',
        'forecast': 'SWPC Forecast F107 data next (3 days)',
        '45day': 'Air Force 45-day Forecast'}

# Dict keyed by inst_id that lists supported tags for each inst_id
inst_ids = {'': [tag for tag in tags.keys() if tag != 'now'], 'obs': ['now'],
            'adj': ['now']}

# Dict keyed by inst_id that lists supported tags and a good day of test data
# generate todays date to support loading forecast data
now = dt.datetime.utcnow()
today = dt.datetime(now.year, now.month, now.day)
tomorrow = today + dt.timedelta(days=1)

# The LASP archive start day is also important
lasp_stime = dt.datetime(1947, 2, 14)

# ----------------------------------------------------------------------------
# Instrument test attributes

_test_dates = {'': {'historic': dt.datetime(2009, 1, 1),
                    'prelim': dt.datetime(2009, 1, 1),
                    'daily': today,
                    'forecast': tomorrow,
                    '45day': today},
               'obs': {'now': dt.datetime(2009, 1, 1)},
               'adj': {'now': dt.datetime(2009, 1, 1)}}

# Other tags assumed to be True
_test_download_ci = {'': {'prelim': False}}

# ----------------------------------------------------------------------------
# Instrument methods

preprocess = methods.general.preprocess


def init(self):
    """Initialize the Instrument object with instrument specific values."""

    # Set the required Instrument attributes
    self.acknowledgements = methods.f107.acknowledgements(self.tag)
    self.references = methods.f107.references(self.tag)
    pysat.logger.info(self.acknowledgements)

    # Define the historic F10.7 starting time
    if self.tag == 'historic':
        self.lasp_stime = lasp_stime

    return


def clean(self):
    """Clean the F10.7 data, empty function as this is not necessary."""

    return


# ----------------------------------------------------------------------------
# Instrument functions


def load(fnames, tag='', inst_id=''):
    """Load F10.7 index files.

    Parameters
    ----------
    fnames : pandas.Series
        Series of filenames.
    tag : str
        Instrument tag. (default='')
    inst_id : str
        Instrument ID, not used. (default='')

    Returns
    -------
    data : pandas.DataFrame
        Object containing satellite data.
    meta : pysat.Meta
        Object containing metadata such as column names and units.

    See Also
    --------
    pysat.instruments.methods.general.load_csv_data

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    # Get the desired file dates and file names from the daily indexed list
    file_dates = list()
    if tag in ['historic', 'prelim', 'now']:
        unique_files = list()
        for fname in fnames:
            file_dates.append(dt.datetime.strptime(fname[-10:], '%Y-%m-%d'))
            if fname[0:-11] not in unique_files:
                unique_files.append(fname[0:-11])
        fnames = unique_files

    # Load the CSV data files
    data = pysat.instruments.methods.general.load_csv_data(
        fnames, read_csv_kwargs={"index_col": 0, "parse_dates": True})

    # Rename the GFZ variable name to be consistent with the other data sets
    if tag == 'now':
        data = data.rename(columns={'F{:s}'.format(inst_id): 'f107'})

    # If there is a date range, downselect here
    if len(file_dates) > 0:
        idx, = np.where((data.index >= min(file_dates))
                        & (data.index < max(file_dates) + dt.timedelta(days=1)))
        data = data.iloc[idx, :]

    # Initialize the metadata
    meta = pysat.Meta()
    desc_prefix = '' if inst_id == '' else '{:s} '.format(inst_id.capitalize())
    meta['f107'] = {meta.labels.units: 'SFU',
                    meta.labels.name: 'F10.7 cm solar index',
                    meta.labels.notes: '',
                    meta.labels.desc:
                    '{:s}F10.7 cm radio flux in Solar Flux Units (SFU)'.format(
                        desc_prefix),
                    meta.labels.fill_val: np.nan,
                    meta.labels.min_val: 0,
                    meta.labels.max_val: np.inf}

    if tag == 'historic':
        # LASP updated file format in June, 2022. Minimize impact downstream by
        # continuing use of `f107` as primary data product.
        if 'f107_adjusted' in data.columns:
            # There may be a mix of old and new data formats.
            if 'f107' in data.columns:
                # Only fill NaN in the `f107` and `f107_adjusted` columns
                # for consistency across both data sets
                data.loc[np.isnan(data['f107']), 'f107'] = data.loc[
                    np.isnan(data['f107']), 'f107_adjusted']

                data.loc[np.isnan(data['f107_adjusted']),
                         'f107_adjusted'] = data.loc[
                             np.isnan(data['f107_adjusted']), 'f107']
            else:
                data['f107'] = data['f107_adjusted']

            # Add metadata
            meta['f107_observed'] = meta['f107']
            raw_str = 'Raw F10.7 cm radio flux in Solar Flux Units (SFU)'
            meta['f107_observed'] = {meta.labels.desc: raw_str}

            meta['f107_adjusted'] = meta['f107_observed']
            norm_str = ''.join(['F10.7 cm radio flux in Solar Flux Units (SFU)',
                                ' normalized to 1-AU'])
            meta['f107_adjusted'] = {meta.labels.desc: norm_str}

            meta['f107'] = {
                meta.labels.desc: meta['f107_adjusted', meta.labels.desc]}

    return data, meta


def list_files(tag='', inst_id='', data_path='', format_str=None):
    """List local F10.7 data files.

    Parameters
    ----------
    tag : str
        Instrument tag, accepts any value from `tags`. (default='')
    inst_id : str
        Instrument ID, not used. (default='')
    data_path : str
        Path to data directory. (default='')
    format_str : str or NoneType
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)

    Returns
    -------
    out_files : pysat._files.Files
        A class containing the verified available files

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    if tag in ['historic', 'now']:
        # Files are by month, going to add date to monthly filename for
        # each day of the month. The load routine will load a month of
        # data and use the appended date to select out appropriate data.
        if format_str is None:
            if tag == 'historic':
                format_str = 'f107_monthly_{year:04d}-{month:02d}.txt'
            else:
                format_str = 'F{:s}_{{year:04d}}-{{month:02d}}.txt'.format(
                    inst_id)
        out_files = pysat.Files.from_os(data_path=data_path,
                                        format_str=format_str)
        if not out_files.empty:
            out_files.loc[out_files.index[-1] + pds.DateOffset(months=1)
                          - pds.DateOffset(days=1)] = out_files.iloc[-1]
            out_files = out_files.asfreq('D', 'pad')
            out_files = out_files + '_' + out_files.index.strftime(
                '%Y-%m-%d')

    elif tag == 'prelim':
        # Files are by year (and quarter)
        if format_str is None:
            format_str = ''.join(['f107_prelim_{year:04d}_{month:02d}',
                                  '_v{version:01d}.txt'])
        out_files = pysat.Files.from_os(data_path=data_path,
                                        format_str=format_str)

        if not out_files.empty:
            # Set each file's valid length at a 1-day resolution
            orig_files = out_files.sort_index().copy()
            new_files = list()

            for orig in orig_files.items():
                # Version determines each file's valid length
                version = np.int64(orig[1].split("_v")[1][0])
                doff = pds.DateOffset(years=1) if version == 2 \
                    else pds.DateOffset(months=3)
                istart = orig[0]
                iend = istart + doff - pds.DateOffset(days=1)

                # Ensure the end time does not extend past the number of
                # possible days included based on the file's download time
                fname = os.path.join(data_path, orig[1])
                dend = dt.datetime.utcfromtimestamp(os.path.getctime(fname))
                dend = dend - pds.DateOffset(days=1)
                if dend < iend:
                    iend = dend

                # Pad the original file index
                out_files.loc[iend] = orig[1]
                out_files = out_files.sort_index()

                # Save the files at a daily cadence over the desired period
                new_files.append(out_files.loc[istart:
                                               iend].asfreq('D', 'pad'))
            # Add the newly indexed files to the file output
            out_files = pds.concat(new_files, sort=True)
            out_files = out_files.dropna()
            out_files = out_files.sort_index()
            out_files = out_files + '_' + out_files.index.strftime('%Y-%m-%d')

    elif tag in ['daily', 'forecast', '45day']:
        out_files = methods.swpc.list_files(name, tag, inst_id, data_path,
                                            format_str=format_str)

    return out_files


def download(date_array, tag, inst_id, data_path, update_files=False,
             mock_download_dir=None):
    """Download F107 index data from the appropriate repository.

    Parameters
    ----------
    date_array : array-like
        Sequence of dates for which files will be downloaded.
    tag : str
        Denotes type of file to load.
    inst_id : str
        Specifies the satellite ID for a constellation.
    data_path : str
        Path to data directory.
    update_files : bool
        Re-download data for files that already exist if True (default=False)
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded (default=None)

    Raises
    ------
    IOError
        If a problem is encountered connecting to the gateway or retrieving
        data from the remote or local repository.

    Warnings
    --------
    Only able to download current forecast data, not archived forecasts.

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """
    # Download standard F107 data
    if tag == 'historic':
        # Test the date array, updating it if necessary
        if date_array.freq != 'MS':
            date_array = pysat.utils.time.create_date_range(
                dt.datetime(date_array[0].year, date_array[0].month, 1),
                date_array[-1], freq='MS')

        # Download from LASP, by month
        freq = pds.DateOffset(months=1, seconds=-1)
        methods.lisird.download(date_array, data_path, 'f107_monthly_',
                                '%Y-%m', 'noaa_radio_flux', freq, update_files,
                                {'f107_adjusted': -99999.0,
                                 'f107_observed': -99999.0},
                                mock_download_dir=mock_download_dir)

    elif tag == 'prelim':
        # Get the local files, to ensure that the version 1 files are
        # downloaded again if more data has been added
        local_files = list_files(tag, inst_id, data_path)

        # Cut the date from the end of the local files
        for i, lfile in enumerate(local_files):
            local_files[i] = lfile[:-11]

        methods.swpc.old_indices_dsd_download(
            name, date_array, data_path, local_files, today,
            mock_download_dir=mock_download_dir)
    elif tag == 'now':
        # Set the download input options
        gfz_data_name = 'F{:s}'.format(inst_id)
        local_file_prefix = '{:s}_'.format(gfz_data_name)

        # Call the download routine
        methods.gfz.json_downloads(date_array, data_path, local_file_prefix,
                                   "%Y-%m", gfz_data_name,
                                   pds.DateOffset(months=1, seconds=-1),
                                   update_files=update_files,
                                   mock_download_dir=mock_download_dir)

    elif tag == 'daily':
        methods.swpc.daily_dsd_download(name, today, data_path,
                                        mock_download_dir=mock_download_dir)

    elif tag == 'forecast':
        methods.swpc.solar_geomag_predictions_download(
            name, date_array, data_path, mock_download_dir=mock_download_dir)

    elif tag == '45day':
        methods.swpc.recent_ap_f107_download(
            name, date_array, data_path, mock_download_dir=mock_download_dir)

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Supports ap index values.

Properties
----------
platform
    'sw'
name
    'ap'
tag
    - 'def' Definitive ap data from GFZ
    - 'now' Nowcast ap data from GFZ
    - 'prediction' Predictions from SWPC for the next 3 days
    - 'forecast' Forecast data from SWPC (next 3 days)
    - 'recent' The last 30 days of Ap data from SWPC
    - '45day' 45-Day forecast data from the Air Force
inst_id
    - ''

Note
----
Downloads data from ftp.gfz-potsdam.de or SWPC. These files also contain Kp
data (with the GFZ data additionally containing Cp data), and so the additional
data files will be saved to the appropriate data directories to avoid multiple
downloads.

The historic definitive and nowcast Kp files are stored in yearly files, with
the current year being updated remotely on a regular basis.  If you are using
historic data for the current year, we recommend re-downloading it before
performing your data processing.

The forecast data is stored by generation date, where each file contains the
forecast for the next three days, the observed daily Ap from the prior day, and
the estimated Ap from the current day. Forecast data downloads are only
supported for the current day. When loading forecast data, the date specified
with the load command is the date the forecast was generated. The data loaded
will span three days. To always ensure you are loading the most recent data,
load the data with tomorrow's date.

Recent data is also stored by the generation date from the SWPC. Each file
contains 30 days of Ap measurements. The load date issued to pysat corresponds
to the generation date.

Examples
--------
::

    ap = pysat.Instrument('sw', 'ap', tag='recent')
    ap.download()
    ap.load(date=ap.tomorrow())


Warnings
--------
The 'forecast', 'recent', and '45day' tags load Ap data for a specific period of
time. Loading multiple files, loading multiple days, the data padding feature,
and multi_file_day feature available from the pyast.Instrument object is not
appropriate for this tag data.

"""

import datetime as dt
import numpy as np
import pandas as pds

import pysat

from pysatSpaceWeather.instruments import methods

# ----------------------------------------------------------------------------
# Instrument attributes

platform = 'sw'
name = 'ap'
tags = {'def': 'Definitive Kp data from GFZ',
        'now': 'Nowcast Kp data from GFZ',
        'prediction': 'SWPC Predictions for the next three days',
        'forecast': 'SWPC Forecast data: prior day to three days hence',
        'recent': 'SWPC provided Kp for past 30 days',
        '45day': 'Air Force 45-day Forecast'}
inst_ids = {'': list(tags.keys())}

# Generate todays date to support loading forecast data
now = dt.datetime.utcnow()
today = dt.datetime(now.year, now.month, now.day)
tomorrow = today + dt.timedelta(days=1)

# ----------------------------------------------------------------------------
# Instrument test attributes

# Set test dates
_test_dates = {'': {'def': dt.datetime(2009, 1, 1),
                    'now': dt.datetime(2020, 1, 1),
                    'forecast': tomorrow,
                    'prediction': tomorrow,
                    'recent': today,
                    '45day': today}}

# ----------------------------------------------------------------------------
# Instrument methods

preprocess = methods.general.preprocess


def init(self):
    """Initialize the Instrument object with instrument specific values."""

    self.acknowledgements = methods.kp_ap.acknowledgements(self.name, self.tag)
    self.references = methods.kp_ap.references(self.name, self.tag)
    pysat.logger.info(self.acknowledgements)
    return


def clean(self):
    """Clean the Kp, not required for this index (empty function)."""

    return


# ----------------------------------------------------------------------------
# Instrument functions


def load(fnames, tag='', inst_id=''):
    """Load Ap index files.

    Parameters
    ----------
    fnames : pandas.Series
        Series of filenames
    tag : str
        Instrument tag (default='')
    inst_id : str
        Instrument ID, not used. (default='')

    Returns
    -------
    result : pandas.DataFrame
        Object containing satellite data
    meta : pysat.Meta
        Object containing metadata such as column names and units

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    meta = pysat.Meta()
    if tag in ['def', 'now']:
        # Load the definitive or nowcast data. The Ap data stored in yearly
        # files, and we need to return data daily.  The daily date is
        # attached to filename.  Parse off the last date, load month of data,
        # and downselect to the desired day
        result = methods.gfz.load_def_now(fnames)

        # Initalize the meta data
        fill_val = np.nan
        for kk in result.keys():
            if kk.lower().find('ap') >= 0:
                methods.kp_ap.initialize_ap_metadata(meta, kk, fill_val)
            elif kk.find('Bartels') >= 0:
                methods.kp_ap.initialize_bartel_metadata(meta, kk)
    else:
        # Load the forecast, recent, prediction, or 45day data
        all_data = []
        for fname in fnames:
            result = pds.read_csv(fname, index_col=0, parse_dates=True)
            all_data.append(result)

        result = pds.concat(all_data)
        fill_val = -1

        # Initalize the meta data
        for kk in result.keys():
            methods.kp_ap.initialize_ap_metadata(meta, kk, fill_val)

    return result, meta


def list_files(tag='', inst_id='', data_path='', format_str=None):
    """List local files for the requested Instrument tag.

    Parameters
    -----------
    tag : str
        Instrument tag, accepts any value from `tags`. (default='')
    inst_id : str
        Instrument ID, not used. (default='')
    data_path : str
        Path to data directory. (default='')
    format_str : str or NoneType
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)

    Returns
    -------
    files : pysat._files.Files
        A class containing the verified available files

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    if tag in ['def', 'now']:
        files = methods.gfz.kp_ap_cp_list_files(name, tag, inst_id, data_path,
                                                format_str=format_str)
    else:
        files = methods.swpc.list_files(name, tag, inst_id, data_path,
                                        format_str=format_str)

    return files


def download(date_array, tag, inst_id, data_path, mock_download_dir=None):
    """Download the Ap index data from the appropriate repository.

    Parameters
    ----------
    date_array : array-like or pandas.DatetimeIndex
        Array-like or index of datetimes to be downloaded.
    tag : str
        Denotes type of file to load.
    inst_id : str
        Specifies the instrument identification, not used.
    data_path : str
        Path to data directory.
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded (default=None)

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    Warnings
    --------
    Only able to download current recent data, not archived forecasts.

    Raises
    ------
    IOError
        If an unknown mock download directory is supplied.

    """

    if tag in ['def', 'now']:
        methods.gfz.kp_ap_cp_download(platform, name, date_array, tag, inst_id,
                                      data_path, mock_download_dir)
    elif tag == 'recent':
        methods.swpc.kp_ap_recent_download(name, date_array, data_path,
                                           mock_download_dir)
    elif tag == 'forecast':
        methods.swpc.geomag_forecast_download(name, date_array, data_path,
                                              mock_download_dir)
    elif tag == 'prediction':
        methods.swpc.solar_geomag_predictions_download(
            name, date_array, data_path, mock_download_dir)
    else:
        methods.swpc.recent_ap_f107_download(name, date_array, data_path,
                                             mock_download_dir)

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Supports the MgII core-to-wing ratio index.

Properties
----------
platform
    'sw'
name
    'mgii'
tag
    - 'composite' Composite data set of MgII core-to-wing index
    - 'sorce' SORCE SOLSTICE MgII core-to-wing index

Examples
--------
Download a month of the composite MgII data and load the second day of the
month.
::

    stime = dt.datetime(1981, 2, 2)
    mgii = pysat.Instrument('sw', 'mgii', tag='composite')
    mgii.download(start=stime)
    mgii.load(date=stime)

"""

import datetime as dt
import numpy as np
import pandas as pds

import pysat

from pysatSpaceWeather.instruments.methods import general
from pysatSpaceWeather.instruments.methods import lisird


# ----------------------------------------------------------------------------
# Instrument attributes

platform = 'sw'
name = 'mgii'
tags = {'composite': 'Composite data set of MgII core-to-wing index',
        'sorce': 'SORCE SOLSTICE MgII core-to-wing index'}

# Dict keyed by inst_id that lists supported tags for each inst_id
inst_ids = {'': [tag for tag in tags.keys()]}

# Dict keyed by inst_id that lists supported tags and a good day of test data
# generate todays date to support loading forecast data
now = dt.datetime.utcnow()
today = dt.datetime(now.year, now.month, now.day)
tomorrow = today + pds.DateOffset(days=1)

# ----------------------------------------------------------------------------
# Instrument test attributes

_test_dates = {'': {'composite': dt.datetime(1981, 11, 6),
                    'sorce': dt.datetime(2005, 3, 6)}}

# ----------------------------------------------------------------------------
# Instrument methods

preprocess = general.preprocess


def init(self):
    """Initialize the Instrument object with instrument specific values."""

    # Set the required Instrument attributes
    self.acknowledgements = lisird.ackn
    self.references = lisird.references(self.platform, self.name, self.tag,
                                        self.inst_id)
    pysat.logger.info(self.acknowledgements)

    return


def clean(self):
    """Clean the MGII data, empty function as this is not necessary."""

    if self.tag == 'sorce':
        pysat.logger.warning(''.join(["The SORCE MGII core-to-wing ratio has ",
                                      "an associated uncertaintly ('unc') ",
                                      "that should be considered when using ",
                                      'the data']))
    return


# ----------------------------------------------------------------------------
# Instrument functions


def load(fnames, tag='', inst_id=''):
    """Load MGII core-to-wing ratio index files.

    Parameters
    ----------
    fnames : pandas.Series
        Series of filenames.
    tag : str
        Instrument tag. (default='')
    inst_id : str
        Instrument ID, not used. (default='')

    Returns
    -------
    data : pandas.DataFrame
        Object containing satellite data.
    meta : pysat.Meta
        Object containing metadata such as column names and units.

    See Also
    --------
    pysat.instruments.methods.general.load_csv_data

    """
    # Get the desired file dates and file names from the daily indexed list
    file_dates = list()
    if tag == 'composite':
        unique_files = list()
        for fname in fnames:
            file_dates.append(dt.datetime.strptime(fname[-10:], '%Y-%m-%d'))
            if fname[0:-11] not in unique_files:
                unique_files.append(fname[0:-11])
        fnames = unique_files

    # Load the CSV data files
    data = pysat.instruments.methods.general.load_csv_data(
        fnames, read_csv_kwargs={"index_col": 0, "parse_dates": True})

    # If there is a date range, downselect here
    if len(file_dates) > 0:
        idx, = np.where((data.index >= min(file_dates))
                        & (data.index < max(file_dates) + dt.timedelta(days=1)))
        data = data.iloc[idx, :]

    # Initalize the metadata
    meta = pysat.Meta()
    notes = {
        'composite': ''.join([
            'The Mg II Index is a proxy for solar chromospheric variability. ',
            'This composite data record is based on the work of Viereck et ',
            'al. (2004) Space Weather, vol 2, CiteID S10005 for measurements',
            ' from 1978 through 2003. For this time range, the Upper ',
            'Atmosphere Research Satellite (UARS) Solar Ultraviolet Spectral ',
            'Irradiance Monitor (SUSIM), UARS Solar Stellar Irradiance ',
            'Comparison Experiment (SOLSTICE), ERS-2/Global Ozone Monitoring ',
            'Experiment (GOME) and five NOAA solar backscatter ultraviolet ',
            'data sets were used. Initially, the best data sets were ',
            'selected. Then the gaps in the record were filled with data ',
            'from various other Mg II data sets. Where no alternate data were',
            ' available, a cubic spline function was used to bridge the ',
            'missing data. In some cases the data gaps were too long for ',
            'reasonable spline fits (more than 5 days), and for these gaps ',
            'the F10.7 cm flux data were scaled to fill the gaps.\nStarting ',
            'in 2003, the data from SORCE SOLSTICE is used exclusively. The ',
            'SOLSTICE spectra have been convolved with a 1.1 nm triangular ',
            'response function to improve the long-term agreement with other ',
            'measurements. All of the data sets have been normalized to a ',
            'commonscale to create a single long-term record.\nFor more ',
            'details please see https://lasp.colorado.edu/home/sorce/',
            'instruments/solstice/solstice-data-product-release-notes/.']),
        'sorce': ''.join([
            'The SORCE Magnesium II core-to-wing index is produced from SORCE',
            ' SOLSTICE instrument measurements using the definition in Snow ',
            'et al. [2005]. The spectral resolution of SORCE SOLSTICE (0.1 ',
            'nm) allows the emission cores of the Mg II doublet to be fully ',
            'resolved and modeled with Gaussian functions. To determine the ',
            'wing irradiances, the spectrum is convolved with a 1.1 nm ',
            'triangular bandpass and then measured at the four wavelengths ',
            'used by NOAA as described in Heath and Schlesinger [1986]. This',
            ' method has several advantages; most importantly it extracts a ',
            'more precise measurement of solar activity [Snow et al., 2005]. ',
            'A simple linear correlation with the standard NOAA data product ',
            'can be used to scale this SORCE measurement to be compatible ',
            'with the long-term composite Mg II index maintained by NOAA.'])}
    meta['mg_index'] = {meta.labels.units: '',
                        meta.labels.name: 'MG II index',
                        meta.labels.notes: notes[tag],
                        meta.labels.desc: 'MG II core-to-wing ratio index',
                        meta.labels.fill_val: np.nan,
                        meta.labels.min_val: 0,
                        meta.labels.max_val: np.inf}

    if 'unc' in data.columns:
        meta['unc'] = {meta.labels.units: '',
                       meta.labels.name: 'MG II Index Uncertainty',
                       meta.labels.notes: "".join([
                           "As described in Snow et al. (2005), the formal ",
                           "uncertainty in each SOLSTICE MgII index ",
                           "measurement is about 0.65%. This is derived from ",
                           "the statistical uncertainty of the underlying ",
                           "spectral measurements and the way they are ",
                           "combined to produce the index."]),
                       meta.labels.desc:
                       'MG II core-to-wing ratio index uncertainty',
                       meta.labels.fill_val: np.nan,
                       meta.labels.min_val: 0,
                       meta.labels.max_val: np.inf}

    return data, meta


def list_files(tag='', inst_id='', data_path='', format_str=None):
    """List local MGII core-to-wing ratio index files.

    Parameters
    ----------
    tag : str
        Instrument tag, accepts any value from `tags`. (default='')
    inst_id : str
        Instrument ID, not used. (default='')
    data_path : str
        Path to data directory. (default='')
    format_str : str or NoneType
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)

    Returns
    -------
    out_files : pysat._files.Files
        A class containing the verified available files

    Note
    ----
    Called by pysat. Not intended for direct use by user.

    """

    # If the format string is not supplied, build the default string
    if format_str is None:
        format_str = "".join([
            "mgii_", tag, "_{year:04d}-{month:02d}",
            '.txt' if tag == 'composite' else '-{day:02d}.txt'])

    # Get the files using the default function
    out_files = pysat.Files.from_os(data_path=data_path, format_str=format_str)

    if tag == 'composite':
        # Files are by month, going to add date to monthly filename for
        # each day of the month. The load routine will load a month of
        # data and use the appended date to select out appropriate data.
        if not out_files.empty:
            out_files.loc[out_files.index[-1] + pds.DateOffset(months=1)
                          - pds.DateOffset(days=1)] = out_files.iloc[-1]
            out_files = out_files.asfreq('D', 'pad')
            out_files = out_files + '_' + out_files.index.strftime(
                '%Y-%m-%d')

    return out_files


def download(date_array, tag, inst_id, data_path, mock_download_dir=None):
    """Download MGII core-to-wing index data from the appropriate repository.

    Parameters
    ----------
    date_array : array-like
        Sequence of dates for which files will be downloaded.
    tag : str
        Denotes type of file to load.
    inst_id : str
        Specifies the ID for the Instrument (not used).
    data_path : str
        Path to data directory.
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date as if they were
        downloaded. (default=None)

    Raises
    ------
    IOError
        If a problem is encountered connecting to the gateway or retrieving
        data from the remote or local repository.

    """

    # Set the download input options
    local_file_prefix = "mgii_{:s}_".format(tag)
    lisird_data_name = "{:s}_mg_index".format(tag)

    if tag == "composite":
        local_file_prefix = "mgii_composite_"
        local_date_fmt = "%Y-%m"
        freq = pds.DateOffset(months=1, seconds=-1)

        # Adjust the date array if monthly downloads are desired
        if date_array.freq != 'MS':
            date_array = pysat.utils.time.create_date_range(
                dt.datetime(date_array[0].year, date_array[0].month, 1),
                date_array[-1], freq='MS')

    elif tag == "sorce":
        local_date_fmt = "%Y-%m-%d"
        freq = dt.timedelta(seconds=86399)

    # Download the desired data
    lisird.download(date_array, data_path, local_file_prefix, local_date_fmt,
                    lisird_data_name, freq, mock_download_dir=mock_download_dir)

    return



from pysatSpaceWeather.instruments import methods  # noqa F401

__all__ = ['ace_epam', 'ace_mag', 'ace_sis', 'ace_swepam', 'sw_ae', 'sw_al',
           'sw_au', 'sw_ap', 'sw_apo', 'sw_cp', 'sw_dst', 'sw_f107', 'sw_flare',
           'sw_hpo', 'sw_kp', 'sw_mgii', 'sw_polarcap', 'sw_sbfield', 'sw_ssn',
           'sw_stormprob']

for inst in __all__:
    exec("from pysatSpaceWeather.instruments import {inst}".format(inst=inst))



"""Core library for pysatSpaceWeather."""

try:
    from importlib import metadata
    from importlib import resources
except ImportError:
    import importlib_metadata as metadata
    import importlib_resources as resources
from os import path

from pysatSpaceWeather import instruments  # noqa F401

__version__ = metadata.version('pysatSpaceWeather')
test_data_path = path.join(
    path.split(str(resources.path(__package__, '__init__.py').__enter__()))[0],
    'tests', 'test_data')



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------

"""Routines for the F10.7 solar index."""

import datetime as dt
import numpy as np
from packaging.version import Version

import pandas as pds
import pysat

import pysatSpaceWeather as pysat_sw


def acknowledgements(tag):
    """Define the acknowledgements for the F10.7 data.

    Parameters
    ----------
    tag : str
        Tag of the space weather index

    Returns
    -------
    ackn : str
        Acknowledgements string associated with the appropriate F10.7 tag.

    """
    lisird = 'NOAA radio flux obtained through LISIRD'
    swpc = ''.join(['Prepared by the U.S. Dept. of Commerce, NOAA, Space ',
                    'Weather Prediction Center'])

    ackn = {'historic': lisird, 'prelim': swpc, 'daily': swpc,
            'forecast': swpc, '45day': swpc,
            'now': pysat_sw.instruments.methods.gfz.ackn}

    return ackn[tag]


def references(tag):
    """Define the references for the F10.7 data.

    Parameters
    ----------
    tag : str
        Instrument tag for the F10.7 data.

    Returns
    -------
    refs : str
        Reference string associated with the appropriate F10.7 tag.

    """
    noaa_desc = ''.join(['Dataset description: ',
                         'https://www.ngdc.noaa.gov/stp/space-weather/',
                         'solar-data/solar-features/solar-radio/noontime-flux',
                         '/penticton/documentation/dataset-description',
                         '_penticton.pdf, accessed Dec 2020'])
    orig_ref = ''.join(["Covington, A.E. (1948), Solar noise observations on",
                        " 10.7 centimetersSolar noise observations on 10.7 ",
                        "centimeters, Proceedings of the IRE, 36(44), ",
                        "p 454-457."])
    swpc_desc = ''.join(['Dataset description: https://www.swpc.noaa.gov/',
                         'sites/default/files/images/u2/Usr_guide.pdf'])
    gfz_desc = ''.join(['Dataset description: https://kp.gfz-potsdam.de/app/',
                        'format/Kp_ap_Ap_SN_F107_format.txt'])

    refs = {'historic': "\n".join([noaa_desc, orig_ref]),
            'prelim': "\n".join([swpc_desc, orig_ref]),
            'now': "\n".join([gfz_desc, orig_ref]),
            'daily': "\n".join([swpc_desc, orig_ref]),
            'forecast': "\n".join([swpc_desc, orig_ref]),
            '45day': "\n".join([swpc_desc, orig_ref])}

    return refs[tag]


def combine_f107(standard_inst, forecast_inst, start=None, stop=None):
    """Combine the output from the measured and forecasted F10.7 sources.

    Parameters
    ----------
    standard_inst : pysat.Instrument or NoneType
        Instrument object containing data for the 'sw' platform, 'f107' name,
        and 'historic', 'prelim', 'now', or 'daily' tag
    forecast_inst : pysat.Instrument or NoneType
        Instrument object containing data for the 'sw' platform, 'f107' name,
        and 'now', 'prelim', '45day' or 'forecast' tag
    start : dt.datetime or NoneType
        Starting time for combining data, or None to use earliest loaded
        date from the pysat Instruments (default=None)
    stop : dt.datetime or NoneType
        Ending time for combining data, or None to use the latest loaded date
        from the pysat Instruments (default=None)

    Returns
    -------
    f107_inst : pysat.Instrument
        Instrument object containing F10.7 observations for the desired period
        of time, merging the standard, 45day, and forecasted values based on
        their reliability

    Raises
    ------
    ValueError
        If appropriate time data is not supplied, or if the date range is badly
        formed.

    Notes
    -----
    Merging prioritizes the standard data, then the 45day data, and finally
    the forecast data

    Will not attempt to download any missing data, but will load data

    """

    # Initialize metadata and flags
    notes = "Combines data from"
    stag = standard_inst.tag if len(standard_inst.tag) > 0 else 'default'
    tag = 'combined_{:s}_{:s}'.format(stag, forecast_inst.tag)
    inst_flag = 'standard'

    # If the start or stop times are not defined, get them from the Instruments
    if start is None:
        stimes = [inst.index.min() for inst in [standard_inst, forecast_inst]
                  if len(inst.index) > 0]
        start = min(stimes) if len(stimes) > 0 else None

    if stop is None:
        stimes = [inst.index.max() for inst in [standard_inst, forecast_inst]
                  if len(inst.index) > 0]
        stop = max(stimes) + pds.DateOffset(days=1) if len(stimes) > 0 else None

    if start is None or stop is None:
        raise ValueError(' '.join(("must either load in Instrument objects or",
                                   "provide starting and ending times")))

    if start >= stop:
        raise ValueError("date range is zero or negative")

    # Initialize the output instrument
    f107_inst = pysat.Instrument()
    f107_inst.inst_module = pysat_sw.instruments.sw_f107
    f107_inst.tag = tag
    f107_inst.date = start
    f107_inst.doy = np.int64(start.strftime("%j"))
    fill_val = None

    f107_times = list()
    f107_values = list()

    # Cycle through the desired time range
    itime = dt.datetime(start.year, start.month, start.day)

    while itime < stop and inst_flag is not None:
        # Load and save the standard data for as many times as possible
        if inst_flag == 'standard':
            # Test to see if data loading is needed
            if not np.any(standard_inst.index == itime):
                # Set the load kwargs, which vary by pysat version and tag
                load_kwargs = {'date': itime}

                # TODO(#131): Remove version check after minimum version
                # supported is 3.2.0
                if all([Version(pysat.__version__) > Version('3.0.1'),
                        Version(pysat.__version__) < Version('3.2.0')]):
                    load_kwargs['use_header'] = True

                if standard_inst.tag == 'daily':
                    # Add 30 days
                    load_kwargs['date'] += dt.timedelta(days=30)

                standard_inst.load(**load_kwargs)

            if standard_inst.empty:
                good_times = [False]
            else:
                good_times = ((standard_inst.index >= itime)
                              & (standard_inst.index < stop))

            if notes.find("standard") < 0:
                notes += " the {:} source ({:} to ".format(inst_flag,
                                                           itime.date())

            if np.any(good_times):
                if fill_val is None:
                    f107_inst.meta = standard_inst.meta
                    fill_val = f107_inst.meta['f107'][
                        f107_inst.meta.labels.fill_val]

                good_vals = standard_inst['f107'][good_times] != fill_val
                new_times = list(standard_inst.index[good_times][good_vals])
                f107_times.extend(new_times)
                new_vals = list(standard_inst['f107'][good_times][good_vals])
                f107_values.extend(new_vals)
                itime = f107_times[-1] + pds.DateOffset(days=1)
            else:
                inst_flag = 'forecast'
                notes += "{:})".format(itime.date())

        # Load and save the forecast data for as many times as possible
        if inst_flag == "forecast":
            # Determine which files should be loaded
            if len(forecast_inst.index) == 0:
                files = np.unique(forecast_inst.files.files[itime:stop])
            else:
                files = [None]  # No load needed, if already initialized

            # Cycle through all possible files of interest, saving relevant
            # data
            for filename in files:
                if filename is not None:
                    load_kwargs = {'fname': filename}

                    # TODO(#131): Remove version check after minimum version
                    # supported is 3.2.0
                    if all([Version(pysat.__version__) > Version('3.0.1'),
                            Version(pysat.__version__) < Version('3.2.0')]):
                        load_kwargs['use_header'] = True

                    forecast_inst.load(**load_kwargs)

                if notes.find("forecast") < 0:
                    notes += " the {:} source ({:} to ".format(inst_flag,
                                                               itime.date())

                # Check in case there was a problem with the standard data
                if fill_val is None:
                    f107_inst.meta = forecast_inst.meta
                    fill_val = f107_inst.meta['f107'][
                        f107_inst.meta.labels.fill_val]

                # Determine which times to save
                if forecast_inst.empty:
                    good_vals = []
                else:
                    good_times = ((forecast_inst.index >= itime)
                                  & (forecast_inst.index < stop))
                    good_vals = forecast_inst['f107'][good_times] != fill_val

                # Save desired data and cycle time
                if len(good_vals) > 0:
                    new_times = list(
                        forecast_inst.index[good_times][good_vals])
                    f107_times.extend(new_times)
                    new_vals = list(
                        forecast_inst['f107'][good_times][good_vals])
                    f107_values.extend(new_vals)
                    itime = f107_times[-1] + pds.DateOffset(days=1)

            notes += "{:})".format(itime.date())

            inst_flag = None

    if inst_flag is not None:
        notes += "{:})".format(itime.date())

    # Determine if the beginning or end of the time series needs to be padded
    if len(f107_times) >= 2:
        freq = pysat.utils.time.calc_freq(f107_times)
    else:
        freq = None
    end_date = stop - pds.DateOffset(days=1)
    date_range = pds.date_range(start=start, end=end_date, freq=freq)

    if len(f107_times) == 0:
        f107_times = date_range

    date_range = pds.date_range(start=start, end=end_date, freq=freq)

    if date_range[0] < f107_times[0]:
        # Extend the time and value arrays from their beginning with fill
        # values
        itime = abs(date_range - f107_times[0]).argmin()
        f107_times.reverse()
        f107_values.reverse()
        extend_times = list(date_range[:itime])
        extend_times.reverse()
        f107_times.extend(extend_times)
        f107_values.extend([fill_val for kk in extend_times])
        f107_times.reverse()
        f107_values.reverse()

    if date_range[-1] > f107_times[-1]:
        # Extend the time and value arrays from their end with fill values
        itime = abs(date_range - f107_times[-1]).argmin() + 1
        extend_times = list(date_range[itime:])
        f107_times.extend(extend_times)
        f107_values.extend([fill_val for kk in extend_times])

    # Save output data
    f107_inst.data = pds.DataFrame(f107_values, columns=['f107'],
                                   index=f107_times)

    # Resample the output data, filling missing values
    if (date_range.shape != f107_inst.index.shape
            or abs(date_range - f107_inst.index).max().total_seconds() > 0.0):
        f107_inst.data = f107_inst.data.resample(freq).fillna(method=None)
        if np.isfinite(fill_val):
            f107_inst.data[np.isnan(f107_inst.data)] = fill_val

    # Update the metadata notes for this procedure
    notes += ", in that order"
    f107_inst.meta['f107'] = {f107_inst.meta.labels.notes: notes}

    return f107_inst


def calc_f107a(f107_inst, f107_name='f107', f107a_name='f107a', min_pnts=41):
    """Calculate the 81 day mean F10.7.

    Parameters
    ----------
    f107_inst : pysat.Instrument
        pysat Instrument holding the F10.7 data
    f107_name : str
        Data column name for the F10.7 data (default='f107')
    f107a_name : str
        Data column name for the F10.7a data (default='f107a')
    min_pnts : int
        Minimum number of points required to calculate an average (default=41)

    Note
    ----
    Will not pad data on its own

    """

    # Test to see that the input data is present
    if f107_name not in f107_inst.data.columns:
        raise ValueError("unknown input data column: " + f107_name)

    # Test to see that the output data does not already exist
    if f107a_name in f107_inst.data.columns:
        raise ValueError("output data column already exists: " + f107a_name)

    if f107_name in f107_inst.meta:
        fill_val = f107_inst.meta[f107_name][f107_inst.meta.labels.fill_val]
    else:
        fill_val = np.nan

    # Calculate the rolling mean.  Since these values are centered but rolling
    # function doesn't allow temporal windows to be calculated this way, create
    # a hack for this.
    #
    # Ensure the data are evenly sampled at a daily frequency, since this is
    # how often F10.7 is calculated.
    f107_fill = f107_inst.data.resample('1D').fillna(method=None)

    # Replace the time index with an ordinal
    time_ind = f107_fill.index
    f107_fill['ord'] = pds.Series([tt.toordinal() for tt in time_ind],
                                  index=time_ind)
    f107_fill.set_index('ord', inplace=True)

    # Calculate the mean
    f107_fill[f107a_name] = f107_fill[f107_name].rolling(window=81,
                                                         min_periods=min_pnts,
                                                         center=True).mean()

    # Replace the ordinal index with the time
    f107_fill['time'] = pds.Series(time_ind, index=f107_fill.index)
    f107_fill.set_index('time', inplace=True)

    # Resample to the original frequency, if it is not equal to 1 day
    freq = pysat.utils.time.calc_freq(f107_inst.index)
    if freq != "86400S":
        # Resample to the desired frequency
        f107_fill = f107_fill.resample(freq).ffill()

        # Save the output in a list
        f107a = list(f107_fill[f107a_name])

        # Fill any dates that fall
        time_ind = pds.date_range(f107_fill.index[0], f107_inst.index[-1],
                                  freq=freq)
        for itime in time_ind[f107_fill.index.shape[0]:]:
            if (itime - f107_fill.index[-1]).total_seconds() < 86400.0:
                f107a.append(f107a[-1])
            else:
                f107a.append(fill_val)

        # Redefine the Series
        f107_fill = pds.DataFrame({f107a_name: f107a}, index=time_ind)

    # There may be missing days in the output data, remove these
    if f107_inst.index.shape < f107_fill.index.shape:
        f107_fill = f107_fill.loc[f107_inst.index]

    # Save the data
    f107_inst[f107a_name] = f107_fill[f107a_name]

    # Update the metadata
    meta_dict = {f107_inst.meta.labels.units: 'SFU',
                 f107_inst.meta.labels.name: 'F10.7a',
                 f107_inst.meta.labels.desc: "81-day centered average of F10.7",
                 f107_inst.meta.labels.min_val: 0.0,
                 f107_inst.meta.labels.max_val: np.nan,
                 f107_inst.meta.labels.fill_val: fill_val,
                 f107_inst.meta.labels.notes:
                 ' '.join(('Calculated using data between',
                           '{:} and {:}'.format(f107_inst.index[0],
                                                f107_inst.index[-1])))}

    f107_inst.meta[f107a_name] = meta_dict

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-.
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.3986138
#
# DISTRIBUTION STATEMENT A: Approved for public release. Distribution is
# unlimited.
# ----------------------------------------------------------------------------
"""Provides support routines for LASP data."""

import datetime as dt
import numpy as np
import os
import pandas as pds
import requests

import pysat


def prediction_downloads(name, tag, data_path, mock_download_dir=None):
    """Download LASP 96-hour prediction data.

    Parameters
    ----------
    name : str
        Instrument name, currently supports Dst, AL, AE, and AU.
    tag : str
        Instrument tag, used to create a descriptive file name.
    data_path : str
        Path to data directory.
    mock_download_dir : str or NoneType
        Local directory with downloaded files or None. If not None, will
        process any files with the correct name and date (following the local
        file prefix and date format) as if they were downloaded (default=None)

    Raises
    ------
    IOError
        If the data link has an unexpected format or an unknown mock download
        directory is supplied.

    """
    times = list()
    data_dict = {name.lower(): []}
    fname = ''.join([name.lower(), '_last_96_hrs.txt'])

    if mock_download_dir is None:
        # Set the remote data variables
        url = ''.join(['https://lasp.colorado.edu/space_weather/dsttemerin/',
                       fname])

        # Download the webpage
        req = requests.get(url)

        # Test to see if the file was found on the server
        if req.text.find('not found on this server') > 0:
            pysat.logger.warning(''.join(['LASP last 96 hour Dst file not ',
                                          'found on server: ', url]))
            raw_txt = None
        else:
            raw_txt = req.text
    else:
        # If a mock download directory was supplied, test to see it exists
        if mock_download_dir is not None:
            if not os.path.isdir(mock_download_dir):
                raise IOError('file location is not a directory: {:}'.format(
                    mock_download_dir))

        # Get the data from the mock download directory
        url = os.path.join(mock_download_dir, fname)
        if os.path.isfile(url):
            with open(url, 'r') as fpin:
                raw_txt = fpin.read()
        else:
            pysat.logger.warning(''.join(['LASP last 96 hour file not found in',
                                          'the local directory: ', url,
                                          ", data may have been saved to an ",
                                          "unexpected filename"]))
            raw_txt = None

    if raw_txt is not None:
        # Split the file into lines, removing the header and
        # trailing empty line
        file_lines = raw_txt.split('\n')[1:-1]

        # Format the data
        for line in file_lines:
            # Split the line on whitespace
            line_cols = line.split()

            if len(line_cols) != 2:
                raise IOError(''.join(['unexpected line encountered in file ',
                                       'retrieved from ', url, ':\n', line]))

            # Format the time and AL values
            times.append(dt.datetime.strptime(line_cols[0], '%Y/%j-%H:%M:%S'))
            data_dict[name.lower()].append(np.float64(line_cols[1]))

    # Re-cast the data as a pandas DataFrame
    data = pds.DataFrame(data_dict, index=times)

    # Write out as a file
    file_base = '_'.join(['sw', name, tag,
                          pysat.utils.time.today().strftime('%Y-%m-%d')])
    file_name = os.path.join(data_path, '{:s}.txt'.format(file_base))
    data.to_csv(file_name)

    return
